<?LassoScript

content_type:'text/xml; charset=UTF8';
	
define_tag:'bail_misformed', -optional='msg';
	fail:-1, 'The POST data was not a valid SOAP request. ' local:'msg';
/define_tag;
protect;

local:'headers'=@client_headers;

if:#headers >> 'SOAPAction:';
	// pop out the operation name as signified by the SOAPAction header
	local:'operation' = (string_findregexp:#headers, -find='(?i)SOAPAction: (.*)\r\n');
	if:#operation->size == 0;
		bail_misformed:'Couldn\'t find SOAPAction header field.';
	/if;
	#operation = #operation->get:2;
	#operation->removeLeading:'"';
	#operation->removeTrailing:'"';
	
	local:'theTag' = @($__SOAP__->find:#operation);

	if:#theTag == null;
		fail:-1, 'The SOAP operation ' #operation ' was not found.';
	/if;
	
	local:'args' = xml:client_postargs;
//	global:'ARGS'=@#args;

    // make a map of the namespaces
	local:'nsmap'=map;

    iterate:$ARGS->extract("//*/namespace::*"), local:'curr'; /* */
        #nsmap->insert(
                #curr->extractOne('normalize-space()') = #curr->extractOne('name()'));
    /iterate;
	
	// SOAP-ENV ns prefix
	var:'SOAPENV' = #nsmap->find('http://schemas.xmlsoap.org/soap/envelope/');
	// LassoService ns prefix
	var:'LS' = #nsmap->find('http://www.blueworld.com/services/LassoService.wsdl');
	// XSI ns prefix
	var:'XSI' = #nsmap->find('http://www.w3.org/1999/XMLSchema-instance');
	// XSD ns prefix
	var:'XSD' = #nsmap->find('http://www.w3.org/1999/XMLSchema');
	// LassoTypes ns prefix
	var:'LPTYPES'=#nsmap->find('http://www.blueworld.com/schemas/lp7types');
	
	// confirm its soapyness
	local:'env' = @#args->(extractOne:"/*[name()='" $SOAPENV ":Envelope']"); /* */
	if: !#env->isa:'xml';
		bail_misformed;
	/if;
	local:'body' = @#env->(extractOne:"*[name()='" $SOAPENV ":Body']");
	if: !#body->isa:'xml';
		bail_misformed:'Couldn\'t find SOAP Body element.';
	/if;
	
	local:'params'=#body->(extract:"*[local-name()='" #operation "' and namespace-uri()='http://www.blueworld.com/services/LassoService.wsdl']/*"); /* */
	log_detail:#params;
	local:'paramsList' = array;
	
	iterate:#params, local:'i';
	   local:'type' = #i->extractOne('normalize-space(attribute::*[name()=\'' $XSI ':type\'])');
	   if:#type->beginsWith($XSD);
            #paramsList->insert(xsd_deserialize(#i, #type->split(':')->last));
	   else;
        	#paramsList->insert:null;
            #paramsList->last->deserialize:#i;
	   /if;
	/iterate;
	
	local:'res' = @#theTag->(run:-params=@#paramsList, -name=#operation);

	local:'ser'=#res->serialize:true;
	
	local:'resultMessage' = 
'<SOAP-ENV:Envelope 
	xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" 
	xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/" 
	xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
	xmlns:xsd="http://www.w3.org/2001/XMLSchema">
	<SOAP-ENV:Body>
		<m:' #operation 'Response xmlns:m="http://www.blueworld.com/services/LassoService.wsdl">
			<m:return xsi:type="' (xsd_getSchemaType:#res->type) '">' #ser '</m:return>
		</m:' #operation 'Response>
	</SOAP-ENV:Body>
</SOAP-ENV:Envelope>';

	$__html_reply__ = #resultMessage;	
else;
	fail:-1, 'Request did not contain the SOAPAction header';
/if;

handle_error;
	// fault
	log_critical:'FAILING';
	
	local:'faultMessage' = 
'<SOAP-ENV:Envelope 
	xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" 
	xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/" 
	xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
	xmlns:xsd="http://www.w3.org/2001/XMLSchema">
	<SOAP-ENV:Body>
		<SOAP-ENV:Fault>
			<faultcode>Client</faultcode>
			<faultstring>' encode_xml:error_msg '</faultstring>
		</SOAP-ENV:Fault>
	</SOAP-ENV:Body>
</SOAP-ENV:Envelope>';
	
	$__html_reply__ = #faultMessage;
/handle_error;
/protect;

?>