<?LassoScript

// set the global types spec
if:(global:'lp7_types.xsd')->type == 'null';
	global:'lp7_types.xsd' = xml:include:lassoapp_link:'XML/lp7types.xsd';
/if;

define_type:'prettyPrintingNSMap', 'map';
    define_tag:'onConvert';
        if (params->first == 'string');
            local:'ret'='';
            iterate:self, local:'i';
                #ret += '<font color="red">' #i->first '</font>';
                #ret += '<br/>\n';
                #ret += '<div class="contentbox">' #i->second '</div>';
            /iterate;
            return:#ret;
        /if;
    /define_tag;
    
    define_tag:'findType', -required='ns', -required='name',
                -description='Returns a type, if found, given a <namespace>:<typename> string.';
        log_detail:'NSMap: ' #ns ' ' #name;
        local:'nsMap' = @self->find(#ns);
        if (#nsMap->isa('map'));
            return(#nsMap->find(#name));
        /if;        
    /define_tag;
/define_type;

define_type:'prettyPrintingTypeMap', 'map';
    define_tag:'onConvert';
        if (params->first == 'string');
            local:'ret'='';
            iterate:self, local:'i';
                #ret += #i->second;
                #ret += '\n';
            /iterate;
            return:#ret;
        /if;
    /define_tag;
/define_type;

define_type:'xsd_attribute';
    local:'name'='';
    local:'base'='';
    local:'use'='';
    local:'fixed'='';
    local:'default'=null;
    local:'comment'='';
    define_tag:'onConvert';
        if (params->first == 'string');
            local:'ret'=('<span class="attributename">'self->name'</span>=<span class="attributebase">'self->base'</span><br/>');
            if (self->comment->size);
                #ret += '<span class="comment">"' self->comment '"</span><br/>';
            /if;
            return:#ret;
        /if;
    /define_tag;    
/define_type;

define_type:'xsd_ref';
    local:'string'='';
    define_tag:'onCreate';
        if (params->size);
            self->string = params->first;
        /if;
    /define_tag;
    define_tag:'onAssign';
        if:params->first->isa:'string';
            self->string = params->first;
            return:true;
        /if;
    /define_tag;
    define_tag:'onConvert';
        if (params->first == 'string');
            return:'<i>' self->string '</i><br/>\n';
        /if;
    /define_tag;
    define_tag:'namespace';
        local:'cpy'= self->string;
        #cpy->removetrailing(self->name);
        return:#cpy;
    /define_tag;
    define_tag:'name';
        return:self->string->split(':')->last;    
    /define_tag;
    define_tag:'size';
        return(self->string->size);
    /define_tag;
    define_tag:'length';
        return(self->string->length);    
    /define_tag;
/define_type;

define_type:'xsd_type';
    local:'name'='';
    local:'namespace'='';
    local:'base'=xsd_ref;
    local:'attributes'=null;
    local:'content'=null;
    local:'contentBinder'='';
    local:'comment'='';

    define_tag:'setName', -required='xml', -required='tns', -required='name';
        if (#name !>> ':');
            self->name = #name;
            self->namespace = #tns;
        else;
            local:'splt'=#name->split(':');
            self->name = #splt->last;
            self->namespace = xsd_lookupNamespaceUri(#xml, #splt->first);
        /if;
    /define_tag;
    define_tag:'onConvert';
        if (params->first == 'string');
            if (self->name->beginsWith:'@');
                local:'p'='<div class="attributegroupbox">';
            else;
                local:'p'='<div class="type">';
            /if;
            #p += '<font color="purple">'self->namespace'</font>:<font color="purple"><b>'self->name'</b></font><br/>\n';
            if (self->base->size);
                #p += self->base;
            else;
                #p += 'NO_BASE';
            /if;
            #p += '<br/>\n';
            if (self->comment->size);
                #p += '<span class="comment">"' self->comment '"</span><br/>';
            /if;
            
            if (self->attributes->size);
                #p += '<div class="attributebox">';                
                iterate:self->attributes, local:'att';
                    #p += #att;
                /iterate;
                #p += '</div>';
            /if;
            if (self->contentBinder);
                #p += self->contentBinder '<br/>';
            else;
                #p += 'NO_CONTENT<br/>';
            /if;
            if (self->content->size);
                #p += '<div class="contentbox">';
                iterate:self->content, local:'c';
                    #p += #c;
                /iterate;
                #p += '</div>';
            /if;
            return:#p + '</div>';
        /if;
    /define_tag;

    define_tag:'make', -required='given', -type='xml', -description='Make a new instance based on this schema info';
        if ($XML_SCHEMA_URIS >> self->namespace);
            if( (:'int', 'integer', 'positiveInteger', 'negativeInteger', 'nonPositiveInteger', 'nonNegativeInteger') >> self->name);
                return:integer(string(#given->contents));
            else((:'double', 'float') >> self->name);
                return:double(string(#given->contents));
            else(self->name == 'string');
                return:string(#given->contents);
            /if;
        /if;
        log_detail:'xsd_type->make: ' self->dump;
        return:string(#given->contents);
        
        /*
        local:'tag' = {   };
        // add the type name
        #tag->addParentType(self->name);        
        local:'made' = #tag->asType;
        if (self->base->size);
            local:'base' = @xsd_generate(self->base->name, self->base->namespace);
            #made->parent->steal(#base);
        /if;
        local:'date' = @#made->properties->first;
        local:'tags' = @#made->properties->second;
        
        return:#made;
        */
    /define_tag;

/define_type;

global:'XML_SCHEMA_URIS' = (:'http://www.w3.org/1999/XMLSchema', 'http://www.w3.org/2001/XMLSchema');
if: (global:'_xml_schema_cache')==null;
    (global:'_xml_schema_cache')->deserialize(include:lassoapp_link:'XML/SOAP/XMLSchemaMap.xml');
/if;

define_tag:'xsd_lookupNamespaceUri', -required='xml', -type='xml', -required='prefix', -type='string';
    return:#xml->extractOne('normalize-space(ancestor-or-self::*/namespace::*[local-name()=\'' #prefix '\'])');
/define_tag;

define_tag:'xsd_fullyQualifyName', -required='xml', -type='xml', 
        -required='tns', -type='string', -required='name', -type='string';
    if (#name !>> ':'); // easy
        return: #tns ':' #name;
    /if;
    local:'splt'=#name->split(':');
    return:#xml->extractOne('normalize-space(ancestor-or-self::*/namespace::*[local-name()=\'' 
                #splt->first 
                '\'])')
                ':' #splt->last;
/define_tag;

define_tag:'xsd_buildType', -required='xml', -type='xml', -required='tns', -type='string';
    select:#xml->name;
        case:'element';
            return:xsd_processElement(#xml, #tns);
        case:'complexType';
            return:xsd_processComplexType(#xml, #tns);
        case:'simpleType';
            return:xsd_processSimpleType(#xml, #tns);
        case:'import';
            return:xsd_processImport(#xml, #tns);
        case:'attribute';
            return:xsd_processAttribute(#xml, #tns);
        case:'attributeGroup';
            return:xsd_processAttributeGroup(#xml, #tns);     
        case:'group';
            return:xsd_processGroup(#xml, #tns);       
        case:'any';
            return:xsd_processAny(#xml, #tns);
        case;
            local:'newt'=xsd_type;
            select:#xml->name;
                case:'choice';
                    xsd_continueChoice(#xml, #tns, #newt);
                case:'sequence';
                    xsd_continueSequence(#xml, #tns, #newt);
                case:'all';
                    xsd_continueAll(#xml, #tns, #newt);
                case:'annotation';
                    xsd_continueAnnotation(#xml, #tns, #newt);
                case;
                    log_warning:'Unhandled XMLSchema node: ' #xml->path;    
            /select;
            return:#newt;
    /select;
/define_tag;

define_tag:'xsd_continueAnyAttribute', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    local('attr' = xsd_attribute);
    local('namespace' = #xml->extract('normalize-space(@namespace)');
    
    select:#namespace;
        case:'##any';
            #attr->base = '*:*';
        case:'##other';
            #attr->base = '!' #tns ':*';
        case:'##targetNamespace';
            #attr->base= #tns ':*';
        case:'##local';
            #attr->base= #tns ':*';            
        case;
            #attr->base = #namespace ':*';
    /select;
    
    #attr->name = '*';
    if (!(#type->attributes->isa:'array'));
        #type->attributes = array;
    /if;
    #type->attributes->insert(@#attr);
/define_tag; 

define_tag:'xsd_continueAttribute', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    local('ref' = #xml->extract('normalize-space(@ref)');
    if (#ref->size);
        if (!(#type->attributes->isa:'array'));
            #type->attributes = array;
        /if;
        #type->attributes->insert(@xsd_ref(xsd_fullyQualifyName(#xml, #tns, #ref)));
        return;
    /if;
    local('attr' = xsd_attribute);
    local('base' = #xml->extract('normalize-space(@type)');
    #attr->base = xsd_fullyQualifyName(#xml, #tns, #base);
    #attr->name = #xml->extract('normalize-space(@name)');
    #attr->use = #xml->extract('normalize-space(@use)');
    #attr->default = #xml->extract('normalize-space(@default)');
    #attr->fixed = #xml->extract('normalize-space(@fixed)');
    if (!(#type->attributes->isa:'array'));
        #type->attributes = array;
    /if;
    #type->attributes->insert(@#attr);
/define_tag;                                        

define_tag:'xsd_continueDocumentation', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type';
    local('uri'=#xml->extract('normalize-space(@source)'));
    if (#uri->size);
        protect;
            #type->comment += '<a href="' #uri '">documentation source</a>';
        /protect;
    /if;
    #type->comment += #xml->contents;                                                                    
/define_tag;   

define_tag:'xsd_continueAnnotation', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type';
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
            case:'documentation';
                xsd_continueDocumentation(#d, #tns, #type);
            case:'appinfo';
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;                              
/define_tag;                                        

define_tag:'xsd_continueExtension', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    local:'base' = #xml->extract('normalize-space(@base)');
    if (!#base);
        log_warning:'Missing base attribute: ' #xml->path;
    else;
        #type->base = xsd_fullyQualifyName(#xml, #tns, #base); // split namespace
    /if;
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
            case:'attribute';
                xsd_continueAttribute(#d, #tns, #type);
            case:'attributeGroup';
                xsd_continueAttributeGroup(#d, #tns, #type);
            case:'choice';
                xsd_continueChoice(#d, #tns, #type);
            case:'group';
                xsd_continueGroup(#d, #tns, #type);
            case:'sequence';
                xsd_continueSequence(#d, #tns, #type);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
/define_tag;                                        

define_tag:'xsd_continueUnion', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    local('memberTypes' = #xml->extract('normalize-space(@memberTypes)'));
    if (#memberTypes->size);
        #type->base = xsd_fullyQualifyName(#xml, #tns, #memberTypes);
    /if;
 
    #type->contentBinder += 'union:';
    #type->content = array;

    iterate: #xml->extract("*"), local:'d';
        select:#d->name;
            case:'simpleType';
                local:'new'=xsd_buildType(#d, #tns);
                if (#new != null);
                    #type->content->insert(#new);
                /if;
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;   
/define_tag;

define_tag:'xsd_continueList', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    local('itemType' = #xml->extract('normalize-space(@itemType)'));
    if (#itemType->size);
        #type->base = xsd_fullyQualifyName(#xml, #tns, #itemType);
    /if;
    #type->contentBinder += 'list:';
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #type);
            case:'simpleType';
                if (#type->content == null);
                    #type->content = array;
                /if;
                #type->content->insert(@xsd_processSimpleType(#d, #tns));
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;    
    
/define_tag;

define_tag:'xsd_continueSimpleContent', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
            case:'restriction';
                xsd_continueRestriction(#d, #tns, #type);
            case:'extension';
                xsd_continueExtension(#d, #tns, #type);
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #type);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;    
/define_tag;

define_tag:'xsd_continueRestriction', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type';
    local('base' = #xml->extract('normalize-space(@base)'));
    if (#base->size);
        #type->base = xsd_fullyQualifyName(#xml, #tns, #base);
    /if;
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
  /*          case:'restriction';
                xsd_continueRestriction(#d, #tns, #type);
            case:'extension';
                xsd_continueExtension(#d, #tns, #type);
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #type);
*/
            case;
 //               log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;    
/define_tag;

define_tag:'xsd_continueComplexContent', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
            case:'extension';
                xsd_continueExtension(#d, #tns, #type);
            case:'restriction';
                xsd_continueRestriction(#d, #tns, #type);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
/define_tag;

define_tag:'xsd_continueAny', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    #type->contentBinder += 'any:';
    #type->content = array;

    iterate: #xml->extract("*"), local:'d';
        local:'new'=xsd_buildType(#d, #tns);
        if (#new != nulls);
            #type->content->insert(#new);
        /if;
    /iterate;    
/define_tag;

define_tag:'xsd_continueAttributeGroup', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    
    local('ref'=#xml->extractOne('normalize-space(@ref)'));
    if (#ref->size);
        if (#type->attributes == null);
            #type->attributes = array;
        /if;
        #type->attributes->insert(@xsd_ref:xsd_fullyQualifyName(#xml, #tns, #ref));
        return;
    /if;   
/*    
    local:'psuedoattr'=xml_type;
    iterate: #xml->extract("*"), local:'d';
        select:#d->name;
            case:'annotation';
                xsd_continueAnnotation(#xml, #tns, #psuedoattr);
            case;
                local:'new'=xsd_buildType(#d, #tns);
                if (#new != null);
                    #psuedoattr->attributes->insert(#new);
                /if;
        /select;
    /iterate;
*/
    #type->attributes->insert(@xsd_processAttributeGroup(#xml, #tns));
/define_tag;

define_tag:'xsd_continueGroup', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    #type->contentBinder += 'group:';
    #type->content = array;

    iterate: #xml->extract("*"), local:'d';
        select:#d->name;
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #type);
            case;
                local:'new'=xsd_buildType(#d, #tns);
                if (#new != null);
                    #type->content->insert(#new);
                /if;
        /select;
    /iterate;
/define_tag;

define_tag:'xsd_continueSequence', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    #type->contentBinder += 'sequence:';
    #type->content = array;

    iterate: #xml->extract("*"), local:'d';
        local:'new'=xsd_buildType(#d, #tns);
        if (#new != null);
            #type->content->insert(#new);
        /if;
    /iterate;
/define_tag;

define_tag:'xsd_continueChoice', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    #type->contentBinder += 'choice:';
    #type->content = array;
    iterate:#xml->extract("*"), local:'d';
        local:'new'=xsd_buildType(#d, #tns);
        if (#new != null);
            #type->content->insert(#new);
        /if;
    /iterate;
/define_tag;

define_tag:'xsd_continueAll', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    #type->contentBinder += 'all:';
    #type->content = array;
    iterate:#xml->extract("*"), local:'d';
        local:'new'=xsd_buildType(#d, #tns);
        if (#new != null);
            #type->content->insert(#new);
        /if;
    /iterate;
/define_tag;

define_tag:'xsd_continueKey', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    // write me!
/define_tag;                                        

define_tag:'xsd_continueSimpleType', -required='xml', -type='xml', -required='tns', -type='string', -required='type';
    iterate:#xml->extract('*'), local:'d';
        select:#d->name;
            case:'restriction';
                xsd_continueRestriction(#d, #tns, #type);
            case:'union';
                xsd_continueUnion(#d, #tns, #type);
            case:'list';
                xsd_continueList(#d, #tns, #type);
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #type);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;    
/define_tag;

define_tag:'xsd_continueComplexType', -required='xml', -type='xml', 
                                        -required='tns', -type='string', 
                                        -required='type', -type='xsd_type';
    local('name' = #xml->extractOne('normalize-space(@name)'));
    if (#name->size);
        #type->setName(#xml, #tns, #name);        
    /if;
    iterate:#xml->extract("*"), local:'d';
        select:#d->name;
            case:'complexContent';
                xsd_continueComplexContent(#d, #tns, #type);
            case:'simpleContent';
                xsd_continueSimpleContent(#d, #tns, #type);
            case:'all';
                xsd_continueAll(#d, #tns, #type);
            case:'sequence';
                xsd_continueSequence(#d, #tns, #type);
            case:'attribute';
                xsd_continueAttribute(#d, #tns, #type);
            case:'choice';
                xsd_continueChoice(#d, #tns, #type);            
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #type);
            case:'group';
                xsd_continueGroup(#d, #tns, #type);
            case:'attributeGroup';
                xsd_continueAttributeGroup(#d, #tns, #type);
            case:'anyAttribute';
                xsd_continueAnyAttribute(#d, #tns, #type);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
    
/define_tag;

define_tag:'xsd_processComplexType', -required='xml', -type='xml', -required='tns', -type='string';
    local:'t'=xsd_type;
    xsd_continueComplexType(#xml, #tns, #t);
    return:#t;
/define_tag;

define_tag:'xsd_processSimpleType', -required='xml', -type='xml', -required='tns', -type='string';
    local('name' = #xml->extractOne('normalize-space(@name)'));
    local:'simpletype' = xsd_type;
	if (#name);
        #simpletype->name = #name;
    /if;
    #simpletype->namespace = #tns;
    xsd_continueSimpleType(#xml, #tns, #simpletype);
    return:#simpleType;     
/define_tag;

define_tag:'xsd_processGroup', -required='xml', -type='xml', -required='tns', -type='string';
    local('name' = #xml->extractOne('normalize-space(@name)'),
            'ref' = #xml->extractOne('normalize-space(@ref)'));
    if (#ref->size);
        return:xsd_ref:xsd_fullyQualifyName(#xml, #tns, #ref);
    /if;    
    local:'element' = xsd_type;
	if (#name);
        #element->name = #name;
    /if;
    #element->namespace = #tns;

    iterate:#xml->extract('*'), local:'d';
        select:#d->name;
            case:'all';
                xsd_continueAll(#d, #tns, #element);
            case:'choice';
                xsd_continueChoice(#d, #tns, #element);
            case:'sequence';
                xsd_continueSequence(#d, #tns, #element);
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #element);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
    return:#element;
/define_tag;

define_tag:'xsd_processAny', -required='xml', -type='xml', -required='tns', -type='string';
    local('namespace' = #xml->extractOne('normalize-space(@namespace)'));
    local:'xsd'=xsd_type;
    #xsd->name = '*';
    select:#namespace;
        case:'##any';
            #xsd->namespace = '*';
        case:'##other';
            #xsd->namespace = '!' #tns;
        case;
            #xsd->namespace = #namespace;
    /select;
    
    iterate:#xml->extract('*'), local:'d';
        select:#d->name;
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #element);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
    return:#xsd;
/define_tag;

define_tag:'xsd_processElement', -required='xml', -type='xml', -required='tns', -type='string';
    local('name' = #xml->extractOne('normalize-space(@name)'),
            'type' = #xml->extractOne('normalize-space(@type)'));

	if(#type->size == 0);
	   #type = #xml->extractOne('normalize-space(@ref)');
	   if (#type->size);
	       // this is a ref
	       // not a named type on its own
	       if (#type !>> ':');
	           #type = #tns ':' #type;
	       else;
	           #type = xsd_fullyQualifyName(#xml, #tns, #type);
	       /if;
	       return:xsd_ref:#type;
	   /if;
	/if;
    local:'element' = xsd_type;
	if (#name);
        #element->name = #name;
    /if;
    #element->namespace = #tns;
	if:#type->size > 0;
        #element->base = xsd_fullyQualifyName(#xml, #tns, #type);
	/if;
	
    iterate:#xml->extract('*'), local:'d';
        select:#d->name;
            case:'complexType';
                xsd_continueComplexType(#d, #tns, #element);
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #element);
            case:'key';
                xsd_continueKey(#d, #tns, #element);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
    return:#element;
/define_tag;

define_tag:'xsd_processAttribute', -required='xml', -type='xml', 
                                        -required='tns', -type='string';
    local('ref' = #xml->extractOne('normalize-space(@ref)');
    if (#ref->size);
        return:xsd_ref:xsd_fullyQualifyName(#xml, #tns, #ref);
    /if;
    local('attr' = xsd_attribute);
    local('base' = #xml->extract('normalize-space(@type)');
    #attr->base = xsd_fullyQualifyName(#xml, #tns, #base);
    #attr->name = #xml->extract('normalize-space(@name)');
    #attr->use = #xml->extract('normalize-space(@use)');
    #attr->default = #xml->extract('normalize-space(@default)');
    #attr->fixed = #xml->extract('normalize-space(@fixed)');
    
    iterate:#xml->extract('*'), local:'d';
        select:#d->name;
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #attr);
            case:'simpleType';
                xsd_continueSimpleType(#d, #tns, #attr);
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;   
    return:#attr;
/define_tag;

define_tag:'xsd_processAttributeGroup', -required='xml', -type='xml', -required='tns', -type='string';
   local('name' = #xml->extractOne('normalize-space(@name)'),
            'type' = #xml->extractOne('normalize-space(@ref)'));
   if (#type->size);
       // this is a ref
       // not a named type on its own
       if (#type !>> ':');
           #type = #tns ':' #type;
       /if;
       return:xsd_ref:#type;
   /if;
    local:'element' = xsd_type;
	if (#name);
        #element->name = '@' + #name; // special attribute group marker
    /if;
    #element->namespace = #tns;
//	#element->contentBinder += 'attributeGroup:';
	#element->content = array;
	#element->attributes = array;
	
    iterate:#xml->extract('*'), local:'d';
        select:#d->name;
            case:'attribute';
                xsd_continueAttribute(#d, #tns, #element);
//            case:'attributeGroup';
//                #element->content->insert(xsd_processAttributeGroup(#d, #tns, #element));
            case:'anyAttribute';
                xsd_continueAnyAttribute(#d, #tns, #element);
            case:'annotation';
                xsd_continueAnnotation(#d, #tns, #element);          
            case;
                log_warning:'Not understood: ' #d->path;
        /select;
    /iterate;
    return:#element;
/define_tag;

define_tag:'xsd_processImport', -required='xml', -type='xml', -required='tns', -type='string';
    local('schemaLocation' = #xml->extractOne("normalize-space(@schemaLocation)"));
    local('namespace' = #xml->extractOne("normalize-space(@namespace)"));
    if (#schemaLocation->size);
        protect;
            local:'d' = include_url:#schemaLocation;
            if (#namespace->size);
                xsd_loadSchema(xml(#d), #namespace);
            else;
                xsd_loadSchema(xml(#d));
            /if;
        /protect;
    /if;
    return:null;
/define_tag;

define_tag:'xsd_processInclude', -required='xml', -type='xml', -required='tns', -type='string';
    local('schemaLocation' = #xml->extractOne("normalize-space(@schemaLocation)"));
    if (#schemaLocation->size);
        protect;
            local:'d' = include_url:#schemaLocation;
            xsd_loadSchema(xml(#d), #tns);
        /protect;
    /if;
    return:null;
/define_tag;

define_tag:'xsd_processSchema', -required='xml', -required='tns';
    local:'tnsmap'=null;
    
    $_xml_schema_cache->insert(#tns=prettyPrintingTypeMap);
    #tnsmap = @$_xml_schema_cache->find(#tns);

    iterate:#xml->extract("*"), local('type');
        if ( $XML_SCHEMA_URIS >> #type->extract('namespace-uri()'));   
            local:'t' = xsd_buildType(#type, #tns);
            if (#t != null);
                #tnsmap->insert( #t->name = #t );
            /if;
        /if;
    /iterate;
/define_tag;

define_tag:'xsd_loadSchema', -required='xml', -type='xml', -optional='inTns', -type='string', 
        -description='Loads and stores information about all contained types.';

    local:'schemas' = #xml->extract("//*[local-name()='schema']"); /* */

    iterate:#schemas, local('schema');
        // check its namespace
        local:'tns' = #schema->extractOne('namespace-uri()');

        if ($XML_SCHEMA_URIS >> #tns);
            if (local_defined('inTns'));
                #tns = #inTns;
            else;
                #tns = #schema->extractOne('normalize-space(@targetNamespace)');
            /if;
            xsd_processSchema(#schema, #tns);
        /if;
    /iterate;
/define_tag;

define_tag:'xsd_getSchemaType', -required='for', -description='Accepts a Lasso type name and returns the name for it in the lp7_types.xsd';
	select:#for;
		case: 'null';
			return:"nullType";	
		case:'string';
			return:"stringType";		
		case:'integer';
			return:"integerType";			
		case:'decimal';
			return:"decimalType";
		case:'bytes';
			return:'bytesType';					
		case:'boolean';			
			return:"booleanType";	
		case:'pair';
			return:"pairType";		
		case:'array';
			return:"arrayType";			
		case:'map';
			return:"mapType";				
		case:'tag';
			return:"customTagType";
		case;
			return:'customType';
	/select;
	fail:-1, 'Unsupported type: ' #for;
/define_tag;
/*
define_tag:'xsd_deserialize', -required='thedata', -type='xml', -required='type', -type='string', 
                    -priority='low', -returntype='integer', 
                    -criteria=((:'int', 'integer', 'positiveInteger', 'negativeInteger', 'nonPositiveInteger', 'nonNegativeInteger') >> #type);
    return:integer(string(#thedata->contents));
/define_tag;

define_tag:'xsd_deserialize', -required='thedata', -type='xml', -required='type', -type='string', 
                    -priority='low', -returntype='double', 
                    -criteria=((:'double', 'float') >> #type);
    return:double(string(#thedata->contents));
/define_tag;

define_tag:'xsd_deserialize', -required='thedata', -type='xml', -required='type', -type='string', 
                    -priority='low', -returntype='string',
                    -criteria=(#type == 'string');
    return:string(#thedata->contents);
/define_tag;
*/
define_tag:'xsd_deserialize', -required='thedata', -type='xml', -priority='low';
    local:'typeName' = #thedata->extractOne("normalize-space(@*[local-name()='type'])");
    log_detail:#typeName;
    if (#typeName->size);
        local:'splt' = #typeName->split(':');
        if (#splt->size > 1);
            local:'namespace' = xsd_lookupNamespaceUri(#thedata, #splt->first);
            local:'name' = #splt->second;
        else;
            local:'namespace' = #thedata->extractOne('normalize-space(namespace-uri(.))');
            local:'name' = #splt->first;
        /if;
    else;
        local:'namespace' = #thedata->extractOne('normalize-space(namespace-uri(.))');
        local:'name' = #thedata->extractOne('normalize-space(name(.))');
    /if;
    // find the namespace map
    local:'xsdtype' = @xsd_cache->findType(#namespace, #name);
    if (#xsdtype != null);
        return:#xsdtype->make(#thedata);
    /if;
    fail:-1, 'Could not find schema for type.';    
/define_tag;
/*
define_tag:'xsd_generate', -required='name', -type='string', -required='namespace', -type='string',
            -description='Generate a blank/default type given the type name and namespace';
    local:'nsmap' = @$_xml_schema_cache->find(#namespace);
    if (#nsmap->isa('map'));
        // find the type name info
        local:'xsdtype'=@#nsmap->find(#name);
        if (#xsdtype != null);
            return:#xsdtype->make;
        /if;
    /if;
    fail:-1, 'Could not find schema for type.';
/define_tag;
*/

define_tag:'xsd_cache';
    return:$_xml_schema_cache;
/define_tag;

?>