<?LassoScript
	
	// Caching Tags
	//
	//	[Cache: 
	//		-Name=(String Required), 
	//		-Expires=(Integer Seconds), 
	//		-Condition=(Boolean Conditional),
	//		-Session=(String Session),
	//		-UseGlobal]
	//	  ... String Contents ...
	//	[/Cache] -> (String Contents)
	//	
	//	[Cache_Object: ..., -Content=(Any Contents)] -> (Any Contents)
	//	
	//	[Cache_Store: ..., -Content=(Any Contents)]
	//
	//	[Cache_Fetch: -Name=..., -Session=..., -UseGlobal] -> (Any Contents)
	//
	//	[Cache_Empty: -Name=..., -Session=..., -UseGlobal]

	define_tag: 'Cache_Object', -priority='replace',
			-required='name', 
			-optional='content',
			-optional='expires', 
			-optional='condition', 
			-optional='session', 
			-optional='useglobal', 
			-optional='key',
			-optional='log',
			-optional='checkonly';
	
		// Load Preferences
		if: !(globals->keys >> '_Cache_Preference') || (globals->(find: '_Cache_Preference')->type != 'map');
			global: '_Cache_Preference' = (_cache_preferences);
		/if;
		local: 'Cache_preference' = globals->(find: '_Cache_Preference');
		local: 'Cache_enable' = #Cache_preference->(find: '_cache_enable');
		local: 'Cache_host' = #Cache_preference->(find: '_cache_host');
		local: 'Cache_expire' = (integer: #Cache_preference->(find: '_cache_expire'));
		local: 'Cache_limit' = (integer: #Cache_preference->(find: '_cache_limit'));

		// Check options
		local: 'c_log' = (local_defined: 'log');
		local: 'c_useglobal' = (local_defined: 'useglobal');
		local: 'c_checkonly' = (local_defined: 'checkonly');
		
		// Bail if caches are disabled
		if: (#Cache_enable == 'N');
			if: #c_checkonly;
				return: true;
			else;
				return: (local: 'content');
			/if;
		/if;
		
		// Check for session and create id
		if: (local_defined: 'session');
			local: 'id' = (session_id: -name=#session);
		else: (#Cache_host == 'Y');
			local: 'id' = server_name;
		else:
			local: 'id' = 'Global';
		/if;
		
		// Append id to cache name
		local: 'c_name' = (string: #name + '|' + #id);
		
		// Connect to cache storage
		if: (#c_useglobal) || !(local_defined: 'session');
			// Initialize the global cache storage
			if: !(globals->keys >> '_Cache_Storage') || (globals->(find: '_Cache_Storage')->type != 'map');
				global: '_Cache_Storage' = (map);
			/if;
			local: 'storage' = @globals->(find: '_Cache_Storage');
		else;
			// Initialize the session cache storage
			local: 'var' = ('_cache_' + #session);
			if: !(var_defined: #var) || ((var: #var)->type != 'map');
				var: #var = (map);
			/if;
			session_addvar: -name=#session, #var;
			local: 'storage' = @(var: #var);
		/if;		
		
		// Find cache record
		local: 'cache' = @(#storage->(find: #c_name));

		// Check whether refresh is needed
		local: 'refresh' = true;
		if: !(local_defined: 'content');
			local: 'refresh' = false;
		else: (#cache === null) || !(#cache->(isa: 'map'));
			// Store the contents in the cache and return it
			local: 'cache' = (map);
			local: 'refresh' = true;
		else: (local_defined: 'condition');
			local: 'refresh' = (boolean: #condition);		
		else: (local_defined: 'expires');
			local: 'c_expires' = (integer: #expires);
			if: (#Cache_expire > 0);
				local: 'c_expires' = (integer: (math_min: #Cache_expire, #c_expires));
			/if;
			local: 'exp' = (date_add: #cache->(find: 'date'), -second=#c_expires);
			local: 'dif' = (date_difference: date, #exp, -secondsbetween);
			local: 'refresh' = (#dif > 0);
		/if;

		if: #c_checkonly;
			return: #refresh;
		/if;

		// Perform refresh if needed and return contents
		if: #refresh;
			#cache->(insert: 'date' = (string: date));
			#cache->(insert: 'type' = #content->type);
			if: (#content->(isa: 'string') || #content->(isa: 'integer') || #content->(isa: 'decimal'));
				#cache->(insert: 'content' = #content);
				#cache->(remove: 'option');
			else;
				#cache->(insert: 'content' = #content->serialize);
				#cache->(insert: 'option' = 'serialize');
			/if;
			if: (local_defined: 'key');
				#cache->(insert: 'content' = (encrypt_blowfish: #cache->(find: 'content'), -seed=(string: #key)));
				if: (#cache->(find: 'option') != null);
					#cache->(insert: 'option' = 'encrypt,' + #cache->(find: 'option'));
				else;
					#cache->(insert: 'option' = 'encrypt');
				/if;
				#cache->(insert: 'key' = (encrypt_md5: (string: #key)));
			/if;
			#cache->(insert: 'id' = #id);
			#cache->(insert: 'stat_fetch' = (integer: #cache->(find: 'stat_fetch')) + 1);
			#cache->(insert: 'stat_store' = (integer: #cache->(find: 'stat_store')) + 1);
			
			#storage->(insert: #c_name = @#cache);
			if: (#c_log == true);
				log_warning: 'Cache: Refreshing ' + #c_name + ' at ' + (string:date) + ' (' + (#content->size) + ').';
			/if;
			return: #content;
		else;
			if: (#c_log == true);
				log_warning: 'Cache: Fetching ' + #c_name + ' from ' + (date: #cache->(find: 'date')) + ' (' + (#cache->(find: 'content')->size) + ').';
			/if;
			#cache->(insert: 'stat_fetch' = (integer: #cache->(find: 'stat_fetch')) + 1);
			if: (#cache->(find: 'option') >> 'encrypt');
				if: (local_defined: 'key') && (#cache->(find: 'key') == (encrypt_md5: (string: #key)));
					local: 'temp' = (decrypt_blowfish: #cache->(find: 'content'), -seed=(string: #key));
				else;
					error_seterrormessage: 'Cache: Invalid key provided for encrypted cache.';
					error_seterrorcode: (-1);
					return: '';
				/if;
			else;
				local: 'temp' = #cache->(find: 'content');
			/if;

			if: (#cache->(find: 'option') >> 'serialize');
				protect;
					#temp->unserialize;
				/protect;
			/if;
			return: #temp;
		/if;
	/define_tag;
	
	define_tag: 'Cache', -container, -priority='replace',
			-required='name', 
			-optional='expires', 
			-optional='condition', 
			-optional='session', 
			-optional='useglobal', 
			-optional='log';
		local: 'params' = (params);
		#params->(removeall: -content);
		#params->(insert: -checkonly);
		#params->(insert: -content='');
		local: 'result' = \cache_object->(run: -params=#params);
		#params->(removeall: -checkonly);
		#params->(removeall: -condition);
		#params->(insert: -condition=#result);
		if: (#result);
			#params->(insert: -content=(string: run_children));
			return: \cache_object->(run: -params=#params);
		else;
			return: \cache_object->(run: -params=#params);
		/if;
	/define_tag;
	
	define_tag: 'Cache_Store', -priority='replace', 
			-required='name', 
			-optional='content',
			-optional='expires', 
			-optional='condition', 
			-optional='session', 
			-optional='useglobal', 
			-optional='log';
		local: 'params' = (params);
		#params->(removeall: -condition);
		#params->(insert: -condition=true);
		\cache_object->(run: -params=#params);
	/define_tag;
	
	define_tag: 'Cache_Fetch', -priority='replace', 
			-required='name', 
			-optional='session', 
			-optional='useglobal', 
			-optional='log';
		local: 'params' = (params);
		#params->(removeall: -condition);
		#params->(removeall: -expires);
		#params->(removeall: -content);
		#params->(insert: -condition=false);
		return: \cache_object->(run: -params=#params);
	/define_tag;
	
	define_tag: 'Cache_Empty', -priority='replace', 
			-required='name', 
			-optional='session',
			-optional='useglobal', 
			-optional='log';
		#params->(removeall: -condition);
		#params->(removeall: -expires);
		#params->(removeall: -content);
		#params->(insert: -condition=true);
		#params->(insert: -expires=0);
		#params->(insert: -content=null);
		\cache_object->(run: -params=#params);
	/define_tag;

	define_tag: '_Cache_Preferences', -privileged;
		local: 'preferences' = (map:
			'_cache_enable' = 'Y',
			'_cache_host' = 'Y',
			'_cache_expire' = (60*60*24),
			'_cache_limit' = (32*1024),
		);
		inline: -search, 
				-database='lasso_internal', 
				-table='global_prefs', 
				-keyfield='store_key', 
				'store_key'='_cache_';
			records;
				#preferences->(insert: (field: 'store_key') = (field: 'data'));
			/records;
		/inline;
		return: @#preferences;
	/define_tag;

?>
