<!--	
	Database Check
	
	This script performs a routine check on all Lasso MySQL databases at startup.
	It runs before LassoStartup.Lasso to ensure that any databases which were not
	closed properly have been repaired if necessary.
	    
-->

<?LassoScript
	{
		log_warning: 'Checking/Repairing Lasso MySQL Databases';
		
		inline: -database='lasso_internal', -sql='SHOW DATABASES';
			records;
				local: 'database' = '`' + (encode_sql: (field: (field_name: 1))) + '`';
				local: 'tables' = (array);
				local: 'repair' = (array);
				inline: -sql=('SHOW TABLES FROM ' + #database);
					records;
						#tables->(insert: #database + '.`' + (encode_sql: (field: (field_name: 1))) + '`');
					/records;
				/inline;

				if: (#tables->size > 0);
					log_warning: '    Checking: ' + #database;
					iterate: #tables, (local: 'table');
						inline: -sql=('CHECK TABLE ' + #table + ' FAST');
							records;
								if: (#repair >> #table) || !((field: 'msg_type') >> 'status');
									log_warning: '        ' (field: 'table') ': ' + (field: 'msg_type') + ' ' + (field: 'msg_text');									
									if: ((field: 'msg_type') >> 'status');
										#repair->(removeall: #table);
									else: !(#repair >> #table);
										#repair->(insert: #table);
									/if;
								/if;
							/records;
						/inline;
					/iterate;
				/if;

				if: (#repair->size > 0);
					Log_Critical: '    Repairing: ' + #database;
					
					iterate: #repair, (local: 'table');
						inline: -sql=('REPAIR TABLE ' + #table);
							records;
								Log_Critical: '        ' (field: 'table') ': ' + (field: 'msg_type') + ' ' + (field: 'msg_text');									
							/records;
						/inline;
					/iterate;
				/if;
				
			/records;
		/inline;
					
		inline: -database='lasso_internal', -sql='SELECT COUNT(*) FROM _errors';
			local: 'error_count' = (integer: (field: (field_name:1)));
			
			if: #error_count > 10000;
			
				log_warning: '    Errors Table: Deleting excess error messages.';
			
				inline: -database='lasso_internal', -sql='DELETE FROM _errors LIMIT ' + (#error_count - 10000);
				/inline;
			
			/if;
			
		/inline;
						
		log_warning: 'Checking/Repairing Done';

	}->(run);
?>
