<!--	[Email_Send] Rev F7a
	Released 16-DEC-2003
	Updates LP 7.0.0
	Maintained by Bil Corry <bil@lasso.biz>

	Many thanks to those on the LassoTalk list for their assistance and feedback,
	this has been a community effort.  Specifically I'd like to thank:

		Chris Balay <List.Admin@mail.curmudgeon-online.com>
		Gunther Bohnert <gunther@gunther-bohnert.de>
		Fabrizio Carioni <fabrizio@golem100.com>
		Steffan A. Cline <steffan@hldns.com>
		Chris Corwin <chris@sonarstudios.com>
		Dawid Czekanowski <blueworld@c3.co.za>
		Ari Denison <pixenter@hotmail.com>
		Steven Q. Evans <sqevans@ucsd.edu>
		Conny Larsson <conny@epicon.se>
		Alan Linnenbank <alinnenbank@yellowmind.com>
		Charles M Neville <charles.neville@cancomuk.com>
		Brian Olsen <brian@qinternet.com>
		Linda Pautsch <ljp@mac.com>
		Alex Pilson <alex@flagshipinteractive.com>
		Adam Randall <adamrandall@bluewoirld.com>
		Fletcher Sandbeck <fletcher@blueworld.com>
		Keith Schuster <keiths@schusterandcompany.com>
		Jeremy Schwartz <jeremy@starmark.com>
		Johan Solve <inbox_js@solve.se>
		Goran Tornquist <goran@3tag.com>
		Tami Williams <tami@asktami.com>


	Still To Do:

		1) Error the email when the attachment can't be found
		2) Add -returnID parameter
		3) Multipart HTML/Text
		4) Implement RetrySeconds
		5) Limit attachments size via preference


-->

<!--	Library of routines to perform low-level email sending via SMTP server
	Implements:
		[Email_Send]	(public routine available to customers)
		[_b_encode_header]
		[_q_encode_header]
		[_qp_encode_body]
		[_SMTP_SendCommand]
		[_SMTP_SendOneMessage]
		[_SMTP_SendOutgoingMail]
		[_SMTP_Init]
		[_SMTP_GetPref]
		[_SMTP_FindEmails]

	Problems:
		Made-up error codes do not fit into overall LDML scheme, I imagine
		Unknown restrictions on what characters may be stored in a SQL database field: do
			customers have to be careful with what body text for their emails is used?  Embedded cr/lf, stuff like that?

	History:
		12/16/2003	FS - Changed to [Net] type and added logging code for debugging.
		11/21/2003	KRJ - Fix quoted-printable encoding so it alwas encodes based on iso8859-1 
						  Insured that the data sent along with commands (RCPT TO, etc.) is always in iso8859-1
		10/22/2003	AJR - Updated the Email_Send tag to work with L7.
		          	      Updated the tag so that it now handles attachments better, and outputs a better version string (for us).
		10/22/2003	AJR - Removed reliance on named_param. Hopefully I didn't break it.
		09/11/2003  FS  - Updated syntax for Lasso Professional 7
		07/10/2003	BC  - Removed charset from the attachments content-disposition as it was tripping up Eudora (thanks Steven Q. Evans)
		06/12/2003	BC  - Now allows email addresses to be separated by semi-colon (;)
		06/12/2003	BC  - Force TCP_Send to retrieve all lines returned, which corrects a problem with Earthlink (thanks Tami Williams)
		06/06/2003	BC  - Added Content-ID to attachment header (thanks Dawid Czekanowski)	
		06/06/2003	BC  - Changed MIME header items to be in order (thanks Dawid Czekanowski)
		06/04/2003	BC  - Changed retry behavior to halt retries only with fatal 5xx errors (thanks Goran Tornquist!)
		06/04/2003	BC  - Fixed octet-stream mime type issue with attachments (thanks Dawid Czekanowski and Linda Pautsch!)
		05/22/2003	BC  - Worked around Lasso timezone issue (thanks Johan Solve!)
		05/07/2003	KRJ - Fixed socket leak issue
		04/30/2003	BC  - Fixed issue with email_send dropping periods (thanks Adam Randall!)
		04/18/2003	BC  - Added SMTP Authentication (RFC 2554)
		03/12/2003	BC  - Fixed -body required bug (thanks Adam Randall!)
		02/24/2003	BC  - Fixed typo with [email_send] declaration (thanks Brian Olsen!)	
		01/08/2003	BC  - Fixed bug in email queue redistribution
		12/17/2002	BC  - Improved error reporting some more, randomized TCP connection name
		12/16/2002	BC  - Improved error reporting, fixed pair bug
		12/16/2002	BC  - Queues log paused/unpaused, fixed bug with redistributing email, dynamcially allows to add additional queues on the fly
		12/13/2002	FS  - Added sleep when queue is paused and between each send batch to prevent processor over-utilization.
		12/13/2002	BC  - fixed attachments bug, updated attachments error trapping
		12/11/2002	FS  - Added default queue number and queue size of 1.
		12/11/2002	FS  - Corrected problem stopping/starting queue (LP6.0.1 Admin.LassoApp also required).
		12/10/2002	BC  - added date/time sent to email, fixed minimum "empty" date to conform to MySQL
		12/09/2002	BC  - corrected some reference issues
		12/03/2002	BC  - corrected some captialization issues (Thanks Fletcher!)
		11/28/2002	BC  - added multiple async queues	
		11/27/2002	BC  - fixed cc bug, added support for non-qp headers, added LIMIT to outgoing email, made email_send -async
		11/05/2002	BC  - optimized email_send tag using iterate
		11/01/2002	BC  - fixed 8bit body bug, added robust error reporting, corrected case on _OutgoingEmail
		08/12/2002	BC  - changed the header "folding" to CRLF<space> from CRLF<tab>.  Some mail servers didn't consider the <tab> a linear whitespace.
		07/31/2002	BC  - implemented Fletcher's base64 encoding routine
		07/30/2002	BC  - changed DateFormat to -DateFormat, reformmated -sql to eliminate whitespace
		07/25/2002	BC  - added support for multiple attachments, attachments column is now of type LONGBLOB
		07/24/2002	BC  - reworked code to now build entire message and store in database in "attachments" field
		07/16/2002	BC  - adjusted code for database tables to change -to, -cc, -bcc columns to type TEXT
							(http://www.listsearch.com/lassotalk.lasso?id=95311)
		07/16/2002	BC  - cleaned up code, removed old unused code
		07/10/2002	BC  - fixed a bug I introduced with -format, it would drop the body
		07/03/2002	BC  - rewrote attachment sending routine
		07/03/2002	BC  - fixed linefeed issue with attachments
		06/28/2002	BC  - added attachment security, can only attach files relative to web root
		06/20/2002	BC  - worked around bug in string->find that would cause improper line wrapping in Q header
		06/17/2002	BC  - the body is now sent as one TCP_Send instead of line by line
		06/17/2002	BC  - rewrote QP wrapping (thx Connie and Johan)
		06/17/2002	BC  - worked around bug in string->find that would cause improper line wrapping in QP body
		06/05/2002	BC  - fixed QP wrapping when body text is less than 75 chars (thx Johan)
		06/05/2002	BC  - added X-Mailer with version for easier debugging of enduser system
		06/04/2002	BC  - fixed Q header encoding to wrap lines longer than 75 chars
		06/03/2002	BC  - fixed \r\n from being broken apart
		05/31/2002	BC  - lots of small fixes, QP Encoding, line returns, etc....
		01/30/2002	GLH - grab name of localhost for logging in to SMTP servers
		01/30/2002	GLH - fix for SIMS servers: use regexp to grab first numeric value, not just space-delimited
		01/17/2002	GLH - prevent bad emails from 'hanging' queue by wrapping more error protection
		01/11/2002	GLH - change outgoing body to conform to RFC \r\n standards for linebreaks
		11/10/2001	GLH - fixed another problem with 'natural' email addresses in the "From" field
		10/24/2001	GLH - fixed bug that caused body not to be retained after message sent
		10/11/2001	GLH - added better email address parser for "Grant Hulbert" <granthulbert@blueworld.com> email addresses
		09/20/2001	GLH - added extra cr/lf between MIME headers and body
		08/21/2001	GLH - Added support for legacy -Email.MIMEType
		08/07/2001	FMS - Changed SMTPerr to blank for successful messages.
		06/20/2001	GLH - changed error reporting to (pair)-based instead of ErrorCode-based, because
					custom tags can't return error codes(!)
		06/18/2001	GLH - added table modification for default email host
					GLH - added default host feature to [Email_Send]
		06/14/2001	GLH - removed after-the-fact table modification to _OutgoingEmailPrefs
		05/11/2001	GLH - added Pause and SleepSeconds prefs
		03/27/2001	GLH - renamed [SendMail] to [Email_Send]
		01/18/2001	GLH - added prefs and initializer routines
		01/17/2001	GLH - removed [Named_Param] custom tag, because it's already defined
		01/12/2001	GLH - streaming -ExtraMIMEHeaders object into database field
		01/11/2001	GLH - Modified Content-Transfer-Encoding to handle HTML emails
		01/10/2001	GLH - Added -ExtraMIMEHeaders support
					GLH - Fixed 76-char linebreaks
					GLH - Fixed ".\r\n" coincidental email data bug
		01/05/2001	GLH - Renamed [SMTP_AppendMail] to simpler [SendMail]
					GLH - Changed String_Concatenate to newer-style " + " coding
					GLH - Added [Named_Param] tag helper routine
					GLH - Updated how we retrieve named parameters, using custom [Named_Param] tag
					GLH - Renamed all private functions with preceding "_", to help with namespace pollution
					GLH - Renamed [_SMTP_SendMail] to simpler [_SMTP_SendOneMessage]
		11/07/2000	GLH - Cosmetic changes to match standard LDML coding style guidelines
					GLH - Added hypen to all email parameters "-Email.to"
					GLH - Renamed Define_Tag functions with _ inserted, like "SMTP_AppendMail"
		11/03/2000	GLH - Debugged some memory leaks from interstitial code (some code was left outside [Define_Tag] Define_Tags)
		11/02/2000	GLH - Added function to append new outgoing email record to database
					GLH - Added function to walk emails database, attempting to send each one
		10/26/2000	GLH - Initial development
-->



[define_tag:'_redistribute_queues', -async, -priority='replace', -privileged]

		<!-- must wait for queues to start up, otherwise will have no queues to assign email to -->
		[sleep: 120000]
		[while: $_SMTP_queue->find:'size' == 0]
			[sleep: 60000]
		[/while]

		[Log_Warning: 'Email: Redistributing Email to Queues']

		<!-- redistribute outgoing emails to queues -->

		<!-- get the number of emails in the queue to divide among all the queues -->
		[Inline:	-Database='site',
				-Maxrecords='all',
				-SQL = 
					"SELECT count(*) as count from _OutgoingEmail " +
					"WHERE " +
						"retries > 0 and " +
						"date_time_sent = '1000-01-01 00:00:00' and " +
						"(left(attachments,3) = 'To:' or " +
						"left(attachments,5) = 'Date:') and " +
						"ExtraMIME is null " +
					"GROUP BY date_time_sent"]
			[local:'queue_chunks' = ((integer: field:'count') / (integer: $_SMTP_queue->find:'size'))]
		[/Inline]

		<!-- divvy up the email to the queues -->
		[loop: ($_SMTP_queue->find:'size')]
			[Inline:	-Database='site',
				-Maxrecords='all',
				-SQL = 
					"UPDATE _OutgoingEmail " +
					"SET ExtraMIME = " + loop_count + " " +
					"WHERE " +
						"retries > 0 and " +
						"date_time_sent = '1000-01-01 00:00:00' and " +
						"(left(attachments,3) = 'To:' or " +
						"left(attachments,5) = 'Date:') and " +
						"ExtraMIME is null " +
					"LIMIT " + #queue_chunks]
			[/Inline]
		[/loop]

		<!-- add any leftovers to queue 1 -->
		[Inline:	-Database='site',
			-Maxrecords='all',
			-SQL = 
				"UPDATE _OutgoingEmail " +
				"SET ExtraMIME = 1 " +
				"WHERE " +
					"retries > 0 and " +
					"date_time_sent = '1000-01-01 00:00:00' and " +
					"(left(attachments,3) = 'To:' or " +
					"left(attachments,5) = 'Date:') and " +
					"ExtraMIME is null "]
		[/Inline]

[/define_tag]


<!--------------------------------------------------------------------->
<!-- _SMTP_Init:
		Creates prefs table and initializes to default values
		Creates _OutgoingEmail table

	Problems:

	Inputs:

	Outputs:
		None

	Side-Effects:
		Creates new _OutgoingEmail table
		Creates new _OutgoingEmailPrefs table
-->
[Define_Tag: '_SMTP_Init', -Priority='Replace']

	[Log_Warning: 'Email: Initializing Queue']

	<!-- Bil Corry: create global variable that contains all ascii hex values that must be QP encoded in message body -->
	[var:'_encodebodychars'=(array:'3D','00','01','02','03','04','05','06','07','08','0B','0C','0E','0F','10','11','12','13','14','15','16','17','18','19','1A','1B','1C','1D','1E','1F','7F','80','81','82','83','84','85','86','87','88','89','8A','8B','8C','8D','8E','8F','90','91','92','93','94','95','96','97','98','99','9A','9B','9C','9D','9E','9F','A0','A1','A2','A3','A4','A5','A6','A7','A8','A9','AA','AB','AC','AD','AE','AF','B0','B1','B2','B3','B4','B5','B6','B7','B8','B9','BA','BB','BC','BD','BE','BF','C0','C1','C2','C3','C4','C5','C6','C7','C8','C9','CA','CB','CC','CD','CE','CF','D0','D1','D2','D3','D4','D5','D6','D7','D8','D9','DA','DB','DC','DD','DE','DF','E0','E1','E2','E3','E4','E5','E6','E7','E8','E9','EA','EB','EC','ED','EE','EF','F0','F1','F2','F3','F4','F5','F6','F7','F8','F9','FA','FB','FC','FD','FE','FF')]

	<!-- Bil Corry: create global variable that contains all ascii hex values that must be Q encoded in message header -->
	[var:'_encodeheaderchars'=(array:'3D','3F','5F','00','01','02','03','04','05','06','07','08','0B','0C','0E','0F','10','11','12','13','14','15','16','17','18','19','1A','1B','1C','1D','1E','1F','7F','80','81','82','83','84','85','86','87','88','89','8A','8B','8C','8D','8E','8F','90','91','92','93','94','95','96','97','98','99','9A','9B','9C','9D','9E','9F','A0','A1','A2','A3','A4','A5','A6','A7','A8','A9','AA','AB','AC','AD','AE','AF','B0','B1','B2','B3','B4','B5','B6','B7','B8','B9','BA','BB','BC','BD','BE','BF','C0','C1','C2','C3','C4','C5','C6','C7','C8','C9','CA','CB','CC','CD','CE','CF','D0','D1','D2','D3','D4','D5','D6','D7','D8','D9','DA','DB','DC','DD','DE','DF','E0','E1','E2','E3','E4','E5','E6','E7','E8','E9','EA','EB','EC','ED','EE','EF','F0','F1','F2','F3','F4','F5','F6','F7','F8','F9','FA','FB','FC','FD','FE','FF')]

	<!-- create table if it doesn't exist already -->
	[Inline: -Database='site',
		-sql='CREATE TABLE IF NOT EXISTS _OutgoingEmailPrefs ' + 
		'( ' +
			'id INT PRIMARY KEY, ' +
			'MaxRetries SMALLINT, ' +
			'RetrySeconds INT, ' +
			'AdminEmail VARCHAR(255), ' +
			'RetainBody SMALLINT, ' +
			'RetainSuccessful SMALLINT, ' +
			'RetainError SMALLINT, ' +
			'DeleteAttachments SMALLINT, ' +
			'TCPTimeout INT, ' +
			'Pause SMALLINT, ' +
			'SleepSeconds INT, ' +
			'DefaultHost VARCHAR(255) ' +
		')']
	[/Inline]

	<!-- If no default prefs record made, then make one -->
	[Inline: -Database='site', -Table='_OutgoingEmailPrefs', -FindAll]
		[If: (Found_Count) == 0]
			[Inline: -Database='site', -Table='_OutgoingEmailPrefs', -Add,
				'id'=				1,
				'MaxRetries'=		5,
				'RetrySeconds'=		30,
				'AdminEmail'=		'',
				'RetainBody'=		0,
				'RetainSuccessful'=	0,
				'RetainError'=		1,
				'DeleteAttachments'=1,
				'TCPTimeout'=		15,
				'Pause'=			0,
				'SleepSeconds'=		15]
			[/Inline]
		[/If]
	[/Inline]

	<!-- check to see if using old tables for email, code provided by Fletcher -->
	[Local: 'table_exists' = false]
	[local: 'table_fields' = (map)]
	[Inline: -Database='site', -SQL='SHOW COLUMNS FROM _OutgoingEmail']
		[if: (found_count) > 0]
			[#table_exists = true]
			[records]
				[#table_fields->(insert: (field: 'field')=(field: 'type'))]
			[/records]
		[/if]
	[/inline]

	[if: #table_exists == true]
		<!-- // Alter table if it does not contain new fields -->
    		[If: (#table_fields->(find: 'toEmail') >> 'VARCHAR')]
			[Inline: -Database='site',
				-SQL='ALTER TABLE _OutgoingEmail CHANGE COLUMN toEmail toEmail TEXT']
			[/Inline]
		[/if]
		[If: (#table_fields->(find: 'cc') >> 'VARCHAR')]
      		[Inline: -Database='site',
        			-SQL='ALTER TABLE _OutgoingEmail CHANGE COLUMN cc cc TEXT']
		 	[/Inline]
		[/if]
		[If: (#table_fields->(find: 'bcc') >> 'VARCHAR')]
			[Inline: -Database='site',
        			-SQL='ALTER TABLE _OutgoingEmail CHANGE COLUMN bcc bcc TEXT']
        		[/Inline]
		[/if]
		[If: (#table_fields->(find: 'Attachments') >> 'VARCHAR')]
			[Inline: -Database='site',
				-SQL='ALTER TABLE _OutgoingEmail CHANGE COLUMN Attachments Attachments LONGBLOB']
			[/Inline]
		[/if]
	[else]

		<!-- create table if it doesn't exist already -->
		[Inline: -Database='site',
		-Sql='CREATE TABLE IF NOT EXISTS _OutgoingEmail ' +
		'( ' +
			'id INT PRIMARY KEY AUTO_INCREMENT, ' +
			'date_time_sub DATETIME, ' +
			'SMTPServer VARCHAR(255), ' +
			'fromEmail VARCHAR(255), ' +
			'toEmail TEXT, ' +
			'subject VARCHAR(255), ' +
			'cc TEXT, ' +
			'bcc TEXT, ' +
			'ExtraMIME TEXT, ' +
			'body LONGBLOB, ' +
			'Attachments LONGBLOB, ' +
			'date_time_sent DATETIME, ' +
			'retries SMALLINT, ' +
			'SMTPErr VARCHAR(255) ' +
			')']
		[/Inline]
	[/if]

	<!-- init outgoing mail queues -->

		<!-- init queue map -->
		[var:'_SMTP_queue' = (map: 'roundrobin' = 1, 'size' = 0)] 	

		<!-- set all outgoing email to null queue for redistribution -->
		[Inline:	-Database='site',
			-Maxrecords='all',
			-SQL = 
				"UPDATE _OutgoingEmail " +
				"SET ExtraMIME = null " +
				"WHERE " +
					"retries > 0 and " +
					"date_time_sent = '1000-01-01 00:00:00' and " +
					"(left(attachments,3) = 'To:' or " +
					"left(attachments,5) = 'Date:')"]
		[/Inline]

		[_redistribute_queues]

[/Define_Tag]


<!--------------------------------------------------------------------->
<!-- Bil Corry: added following custom tag, not used but left it in, maybe future use? -->
<!-- _b_encode_header:
		Returns a B (BASE64) encoded string for message header, given a string

	Problems:
		If longer than 75 chars, then violates RFC 1522

	Inputs:
		String like: To: Bil Corry <bil@bilcorry.com>

	Outputs:
		BASE64 encoded string

	Side-Effects:

-->
[define_tag:'_b_encode_header', -Priority='Replace']
	[local:'raw'=params->(get:1)]
	[#raw = '=?iso-8859-1?B?' + (encode_base64:#raw) + '?=']
	[return:#raw]
[/define_tag]


<!--------------------------------------------------------------------->
<!-- Bil Corry: added following custom tag, written by Johan Solve -->
<!-- _q_encode_header:
		Returns an Q (quoted-printable) encoded string for message header, given a string

	Problems:

	Inputs:
		String like: To: Bil Corry <bil@bilcorry.com>

	Outputs:
		Q (quoted-printable) encoded string

	Side-Effects:
		Relys on $_encodeheaderchars being globally declared in [_STMP_Init]

-->

[define_tag:'_q_encode_header', -Priority='Replace', -required='raw_ref',-required='charset']
<!-- per RFC 1522 http://www.ietf.org/rfc/rfc1522.txt -->

	[if: #raw_ref == ''][return: ''][/if]

	[local:'raw' = #raw_ref] <!-- detact the reference -->
	[local:'original'=#raw]

	<!-- encode all non-vanilla ascii characters -->
	<!-- have to do this as binary data -->
	[local:'whatswiththesquarebrackets'=bytes];
	[#whatswiththesquarebrackets->(importAs:#raw, 'iso8859-1')]
	[iterate: $_encodeheaderchars, local:'encodechar']
		[#whatswiththesquarebrackets = (string_replaceregexp: #whatswiththesquarebrackets,-find ='\\x' + #encodechar,-replace = '=' + #encodechar)]
	[/iterate]
	[#raw = #whatswiththesquarebrackets->exportString:'iso8859-1']

	<!-- check if any encoding was done -->
	[if: #original == #raw]
		[return:#original]
	[else]
		<!-- encode spaces as underscores -->
		[#raw = (string_replace: #raw,-find=' ',-replace='_')]

		<!-- check line length, can not exceed 75 chars per line per RFC 1522 -->
		<!-- that includes all formatting for Q format, which adds 17 more chars -->
		<!-- so actual length per line is limited to 58 chars without formatting -->

		<!-- loop through and break header into 58 char lines, then add formatting -->
		[local: 'encoded'=NULL]
		[local: 'TotalLen'= #raw->length ]
		[local: 'Start'=1]

		[While: #TotalLen > 0]

			<!-- add RFC separator CRLF SPACE between encoded words -->
			[if: #Start != 1]
				[#encoded += '\r\n ']
			[/if]

			<!-- to make sure we don't break apart Q-encoded char, such as =FF, -->
			<!-- we shorten the line length -->
			<!-- Bil Corry: 20-JUN-2002 implemented string->find bug workaround -->
			[if: #TotalLen > 58]
				[Local: 'LineLength' = 58 ]
				[if: #raw->(substring:#Start + 56,2)->(find:'=') > 0]
					[Local: 'LineLength' = 55 + #raw->(substring:#Start + 56,2)->(find:'=')]
				[/if]
			[else]
				[Local: 'LineLength' = #TotalLen ]
			[/if]

			<!-- put into the Q format and add to encoded value to be returned -->
			[#encoded += '=?'+#charset+'?Q?' + #raw->(substring: #Start, #LineLength) + '?=']

			[Local: 'Start'= #Start + #LineLength]
			[Local: 'TotalLen'= #TotalLen - #LineLength]
		[/While]

		[return:#encoded]
	[/if]
[/define_tag]


<!--------------------------------------------------------------------->
<!-- Bil Corry: added following custom tag, written by Johan Solve -->
<!-- _qp_encode_body:
		Returns an quoted-printable encoded string for message body, given a string

	Problems:

	Inputs:
		String like: "Grant Hulbert" <granthulbert@blueworld.com>, <fred@fred.com>

	Outputs:
		qp-encoded string

	Side-Effects:
		Relys on $_encodebodychars being globally declared in [_STMP_Init]

-->
[define_tag:'_qp_encode_body', -Priority='Replace', -required='raw_ref']
	[if: #raw_ref == ''][return: ''][/if]
	[local:'raw' = bytes];
	[#raw->(importAs:#raw_ref, 'iso8859-1')]
	[iterate: $_encodebodychars, local:'encodechar']
		[#raw = (string_replaceregexp: #raw,-find ='\\x' + #encodechar,-replace = '=' + #encodechar)]
	[/iterate]
	[return:#raw->exportString:'iso8859-1']
[/define_tag]


<!--------------------------------------------------------------------->
<!-- _SMTP_FindEmails:
		Returns an array of all valid email addresses, given a string

	Problems:

	Inputs:
		String like: "Grant Hulbert" <granthulbert@blueworld.com>, <fred@fred.com>

	Outputs:
		array of cleaned email addresses

	Side-Effects:
-->
[define_tag: '_SMTP_FindEmails', -Priority='Replace']
	[Local: 'emailArray' = (String_Replace: params->(get: 1), -find=';', -replace=',')]
	[Local: 'emailArray' = (String_FindRegExp: #emailArray, -find='[^ @\r\n,]*@[^ @\r\n,]*')]
	[Local: 'outArray' = (array)]
	[Loop: #emailArray->size]
		[If: #emailArray->(get: loop_Count)->(get: 1,1) == '<']
			[#outArray->(insert: (String_FindRegExp: #emailArray->(get: Loop_Count), -Search='<([^@]*@[^@]*)>')->get:2)]
		[else]
			[If: #emailArray->(get: loop_Count)->(get: 1,1) != '"']
				[#outArray->(insert: #emailArray->(get: Loop_Count))]
			[/If]
		[/If]
	[/Loop]
	[return: #outArray]
[/define_tag]


<!--------------------------------------------------------------------->
<!-- _SMTP_GetPref:
		Returns the value of an email preference

	Problems:
		No error-checking if pref name requested is undefined

	Inputs:
		*unnamed*	name of preference, one of the following:
						id
						MaxRetries
						RetrySeconds
						AdminEmail
						RetainBody
						RetainSuccessful
						RetainError
						DeleteAttachments

	Outputs:
		value of preference

	Side-Effects:
-->
[Define_Tag: '_SMTP_GetPref', -Priority='Replace', -privileged]
	[If: ((Global_Defined: '__SMTP_Preferences__') == false) || ((Global: '__SMTP_Preferences__')->Type != 'map')]
		[Global: '__SMTP_Preferences__' = (map)]
		[Inline: -Database='site', -Table='_OutgoingEmailPrefs', -FindAll]
			[Loop: (Field_Name: -Count)]
				[$__SMTP_Preferences__->(Insert: (Field_Name: Loop_Count)=(Field: (Field_Name: Loop_Count)))]
			[/Loop]
		[/Inline]
	[/If]
	[Return: $__SMTP_Preferences__->(Find: (params->get:1))]
[/Define_Tag]


<!--------------------------------------------------------------------->
<!-- Email_Send:
		Queues an email for later sending.  Roughly equivalent to inline version of email sending

	Problems:
		No error checking to make sure all required params are supplied
		-Attachments implies more than one file, but we only support 1 right now

	Inputs:
		-Host			(optional ) ip address or name of SMTP server
		-From			email address of sender
		-To				email address of recipient
		-Subject		(optional) subject line in outgoing email
		-CC				(optional) email address of carbon-copy recipient
		-BCC			(optional) email address of blind-carbon-copy recipient
		-Body			(optional) text of message to be sent
		-Format			(optional) file where text of message resides
		-Attachments	(optional) Name of file to attach to outgoing message
		-ExtraMIMEHeader	(optional) Array of pairs that go into extra MIME headers

	Outputs:
		None

	Side-Effects:
		Appends a record to site._OutgoingEmail that contains all info required to send the email
-->
[Define_Tag: 'Email_Send',
	-privileged,
	-Priority='Replace',
	-required='to',
	-required='from',
	-required='subject',
	-optional='body',
	-optional='cc',
	-optional='bcc',
	-optional='attachments',
	-optional='extramimeheaders',
	-optional='format',
	-optional='host',
	-optional='mimetype']

	<!-- init message body -->
	[local:'_message'='']
	[local: '_body'=(string)]
	[local:'_content_boundary'='***************_multipart_boundary_***************']

	<!-- init body if wasn't passed -->
	[if: (local_defined: 'body')]
		[#_body = #body]

	<!-- figure out if the body of the message comes from disk or is passed as text -->
	[else: (local_defined: 'Format')]
		[#_body=(Include: #format)]
	[else]
		[#_body = '\n']
	[/If]

	<!-- Use default host if none provided -->
	[If: !(local_defined: 'Host')]
		<!-- PROBLEM: default host still can't be read from prefs yet -->
		[local:'Host'=(_LLGetDefaultHost)]
	[/If]

	<!-- clean up cr/lf in outgoing email body to conform to email RFC standard \r\n -->
	[#_body = 	(string_replace:
					(string_replace:
						(string_replace:
							#_body,
							-find='\r\n',
							-replace='\r'),
						-find='\n',
						-replace='\r'),
					-find='\r',
					-replace='\r\n')
	]

	<!-- work around an array copy bug -->
	[local: '_attachments' = (local: 'attachments')]

	<!-- get attachment name(s), convert to array if just single string -->
	[if: (local_defined: 'attachments')]
		[if: #_attachments->type != 'array']
				[if:#_attachments->size == 0]
					[#_attachments = (array)]
				[else]
				[#_attachments = (array:#_attachments)]
			[/if]
		[/if]
	[/if]

	[If: !(local_defined: 'MIMEType')]
		[local: 'mimetype' = 'text/plain']
	[/If]
	
	[Local: '-ExtraMIMEHeaders'=(Array:'MIME-Version'='1.0',
								'Content-Type'= #mimetype + ';\r\n charset="iso-8859-1"',
								'Content-Transfer-Encoding'='quoted-printable')]


	<!-- if MIMEs passed in, then use them -->
	[If: (local_defined: 'ExtraMIMEHeaders')]
		[if: #ExtraMIMEHeaders->type == 'array' || #ExtraMIMEHeaders->type == 'map']
			[iterate: #ExtraMIMEHeaders,local:'item']
				[if: #item->type == 'pair']
					[if: #-ExtraMIMEHeaders->find:(#item->first)->size > 0]
						[#-ExtraMIMEHeaders->get:(#-ExtraMIMEHeaders->(findindex:(#item->first))->get:1)->second = #item->second]
					[else]
						[#-ExtraMIMEHeaders->(insert: #item->first = #item->second)]
					[/if]
				[else]
					[#-ExtraMIMEHeaders->(insert: 'X-Debug-MIMEItem' = 'MIME Pair Not Passed')]
				[/if]
			[/iterate]
		[else]
			[#-ExtraMIMEHeaders->(insert: 'X-Debug-ExtraMIMEHeaders' = 'Not Passed as Array')]
		[/if]
	[/If]

	<!-- check for CC and BCC -->
	[if: !(local_defined: 'CC')]
		[local:'CC'='']
	[/if]
	[if: !(local_defined: 'BCC')]
		[local:'BCC'='']
	[/if]

	<!-- get charset -->
	[if: (#-ExtraMIMEHeaders->find:'Content-Type'->get:1->second) >> 'charset=']
		[local:'charset' = (#-ExtraMIMEHeaders->find:'Content-Type'->get:1->second)->split:'='->last]
		[#charset = (string_replace: #charset, -find='"',-replace='')]
		[#charset = (string_replace: #charset, -find="'",-replace='')]
	[else]
		[local:'charset'='iso-8859-1']
	[/if]

	<!-- work around Lasso timezone issue discovered by Johan -->
	[local:'timezone' = (date_getcurrentdate)->(format:'%z')]
	[if: !(#timezone->beginswith:'-') && !(#timezone->beginswith:'+')]
		[#timezone = '+' + #timezone]
	[/if]
	
	<!-- send "Date: Mon, 2 Oct 1981 15:33:22 -0800" -->
	[#_message += 'Date: ' + ((date_getcurrentdate)->(format:'%a, %-d %b %Y %T ')) + #timezone + '\r\n']

	<!-- check if we QP the headers -->
	[if: #-ExtraMIMEHeaders->find:'Content-Transfer-Encoding'->get:1->second == 'quoted-printable' || #charset == 'iso-8859-1']
	
		<!-- send "To: you@yourdomain.com\r\n" MIME header-->
		[#_message += 'To: ' + (_q_encode_header: -raw_ref=#to, -charset=#charset) + '\r\n']

		<!-- send "From: me@mydomain.com\r\n" MIME header -->
		[#_message += 'From: ' + (_q_encode_header: -raw_ref=#from, -charset=#charset) + '\r\n']

		<!-- send "Subject: test\r\n" MIME header -->
		[#_message += 'Subject: ' + (_q_encode_header: -raw_ref=#subject, -charset=#charset) + '\r\n']

		<!-- send "Cc: me@mydomain.com\r\n" MIME header -->
		[If: (local: 'cc')->size > 0]
			[#_message += 'Cc: ' + (_q_encode_header: -raw_ref=#cc, -charset=#charset) + '\r\n']
		[/If]

	[else]
	
		<!-- send "To: you@yourdomain.com\r\n" MIME header-->
		[#_message += 'To: ' + #to + '\r\n']

		<!-- send "From: me@mydomain.com\r\n" MIME header -->
		[#_message += 'From: ' + #from + '\r\n']

		<!-- send "Subject: test\r\n" MIME header -->
		[#_message += 'Subject: ' + #subject + '\r\n']

		<!-- send "Cc: me@mydomain.com\r\n" MIME header -->
		[If: (local: 'cc')->size > 0]
			[#_message += 'Cc: ' + #cc + '\r\n']
		[/If]	
	
	[/if]

	[#_message += 'X-Mailer: Lasso Professional ' + (string_replaceregexp: -find = '([\\.\\d]+).*', -replace = '\\1', (lasso_version: -lassoversion)) + '\r\n']

	<!-- Send optional attachment headers -->
	[If: #_attachments->size > 0]
			[#_message += 'MIME-Version: 1.0\r\n']
			[#_message += 'Content-Type: multipart/mixed;\r\n boundary="'+#_content_boundary+'"\r\n\r\n']
			<!-- Bil Corry: added preamble to mime message, thx to Jeremy Schwartz for the verbage -->
			[#_message += 'This message is in MIME format. Since your mail reader does not understand this format, some or all of this message may not be legible.\r\n']

			<!-- Add content boundaries iff we're sending attachments -->
			[#_message += '\r\n--'+#_content_boundary+'\r\n']
	[/if]

	<!-- Send MIME headers -->
		[iterate: #-ExtraMIMEHeaders,local:'item']
			[#_message += (#item->first) + ': ' + (#item->second) + '\r\n']
		[/iterate]
		[#_message += '\r\n']

	[If: #-ExtraMIMEHeaders->find:'Content-Transfer-Encoding'->get:1->second == 'quoted-printable']

		<!-- QP encode theBody -->
		[Local: 'theBody' = (_qp_encode_body: -raw_ref=#_body)]

		<!-- check that we have a body! -->
		[if:#theBody->length > 0]

			<!-- format it into 76-character chunks for transmission -->
			<!-- what we're trying to do is make sure it doesn't exceed 76-chars -->
			<!-- and will use the body's hardwraps when we can and only break up long lines -->
			[local:'theBodyChunks'=#theBody->(split:'\r\n')]

			<!-- the array that will hold the body to be sent -->
			[local:'theBodyOut'=(array)]

			<!-- check each chunk for size -->
			[iterate:#theBodyChunks,local:'chunk']

				<!-- check for period at start of line -->
				[if: #chunk->length > 0]
					[if: #chunk->get:1 == '.']
						[#chunk = '.' + #chunk]
					[/if]
				[/if]

				<!-- ensure each chunk is max 76 chars -->
				[while: #chunk->length > 76]

					<!-- line size is 75 chars plus softwrap char '=' for total of 76 chars -->
					[local:'lineSize'=75]

					<!-- grab the first 75 chars, we'll add '=' to end of line -->
					[local:'temp'= #chunk->(substring:1,#lineSize)]

					<!-- check that we're not breaking up QP-encoded letters such as =FF -->
					[if: #temp->(find:'=',74) > 0]

						<!-- adjust line size -->
						[#lineSize = #temp->(find:'=',74) - 1]

						<!-- get the line -->
						[local:'temp'= #chunk->(substring:1, #lineSize)]

					[/if]

					<!-- soft wrap the line with QP-code -->
					[#temp += '=']

					<!-- build the array for transmission -->
					[#theBodyOut->(insert:#temp)]

					<!-- get next section of the "chunk" -->
					[#chunk = #chunk->(substring:#lineSize + 1)]

					<!-- check for period at start of line -->
					[if: #chunk->length > 0]
						[if: #chunk->get:1 == '.']
							[#chunk = '.' + #chunk]
						[/if]
					[/if]
												
				[/while] <!-- ensure each chunk is max 76 chars -->

				<!-- build the array for transmission -->
				[#theBodyOut->(insert:#chunk)]

			[/iterate]

			<!-- build the body for transmission -->
			[#theBody = null]
			[iterate:#theBodyOut,local:'item']
				[#theBody += #item + '\r\n']
			[/iterate]

			<!-- send the body -->
			[#_message += #theBody]

		[/if] <!-- do we have a body? -->

	[Else]
		[#_message += #_body]
	[/If]

	<!-- Send actual attachment body -->
	[iterate: #_attachments,local:'oneAttachment']

		[local: 'path_split'=#oneAttachment->(split: '/')]
		[local: 'filename'=#path_split->(get: #path_split->size)]
	
		[local: 'file_error'='']
		[local: 'file_raw'='']
		[protect]
			[#file_raw=(encode_base64:(include_raw: #oneAttachment))]
			[if: error_currenterror != error_noerror]
				[#file_error = 'Fatal error attaching file: ' + error_currenterror]
			[/if]
			[if: file_currenterror != error_noerror]
				[#file_error = 'File error attaching file: ' + file_currenterror]
			[/if]						
			[handle: error_currenterror != error_noerror || file_currenterror != error_noerror]
				[if: file_currenterror != error_noerror]
					[#file_error = 'File error attaching file: ' + file_currenterror]
				[else]
					[#file_error = 'Fatal error attaching file: ' + error_currenterror]
				[/if]
			[/handle]
		[/protect]

		[if: #file_error != '']
			[#filename = 'file_error_'+loop_count+'.txt']
			[#file_raw = (encode_base64: #file_error)]
		[/if]
	
		[local: 'file_out'='']
		[local: 'index' = 1]
		[local: 'file_raw_size'=#file_raw->length]
		[while: #index <= #file_raw_size]
			[#file_out += #file_raw->(substring: #index, 76) + '\r\n']
			[#index += 76]
		[/while]
	
		[#_message += '\r\n--'+#_content_boundary+'\r\n']
		[#_message += 'Content-Type: application/octet-stream;\r\n name="' + #filename + '"\r\n']
		[#_message += 'Content-Transfer-Encoding: base64\r\n']
		[#_message += 'Content-Disposition: attachment;\r\n filename="' + #filename + '"\r\n']
		[#_message += 'Content-ID: <' + #filename + '>\r\n']

<!--
		[#_message += 'X-Debug-Index: '+#index+'\r\n']
		[#_message += 'X-Debug-Size: '+#file_raw_size+'\r\n']
		[#_message += 'X-Debug-Name: '+#oneAttachment+'\r\n']
-->
		[#_message += '\r\n']

		<!-- now send the entire base64 file -->
		[#_message += #file_out ]

	[/iterate]
	
	[if: #_attachments->size > 0]
		<!-- closing content boundary -->
		[#_message += '\r\n--'+#_content_boundary+'--\r\n']
	[/if]

	<!-- get queue for round robin -->
	[local:'queue_robin' = $_SMTP_queue->find:'roundrobin']

	<!-- PROBLEM: maxretries still not read from db: (_SMTP_GetPref: 'MaxRetries') -->

	[Inline: -Add,
			-Database='site',
			-Table='_OutgoingEmail',
			'date_time_sub' = Date->(Format: '%Q %T'),
			'SMTPServer' = #host,
			'fromEmail' = #from,
			'toEmail' =	#to,
			'subject' =	#subject,
			'cc' = #CC,
			'bcc' = #BCC,
			'body' = #_body,
			'ExtraMIME' = #queue_robin,
			'attachments' =	#_message,
			'date_time_sent' = '1000-01-01 00:00:00',
			'retries'= (_SMTP_GetPref: 'MaxRetries'),
			'SMTPErr'= '']
	[/inline]

	<!-- update queue for round robin -->
	[#queue_robin += 1]
	<!-- check for rollover -->
	[if: #queue_robin > ($_SMTP_queue->find:'size')]
		[#queue_robin = 1]
	[/if]
	[$_SMTP_queue->(insert: 'roundrobin' = #queue_robin)]
	
[/Define_Tag]

<!--------------------------------------------------------------------->
<!-- _SMTP_SendCommand:
		Sends an indiviual command line to SMTP server, as part of a more-complete
			dialog between client and server.

	Problems:
		made-up error code of -8001

	Inputs:
		-Text		body of text to send out on TCP stream
		-Timeout	seconds to wait before giving up
		-GoodAnswer	text of error code that indicates a valid response (220 or 250 for SMTP)

	Outputs:
		Returns a pair: error code + text of message from SMTP server, like "250 2.1.0 ... Sender ok"
		Sets error message to returned string from mail server, like "250 2.1.0 ... Sender ok"
		Error code will be 0 if a good response was received from SMTP server

	Side-Effects:
		Sends text out on the #TCPName named TCP stream
		If there was an error on the TCP stream itself, then Error_CurrentError is set but no output is returned
		If returned string from mail server is an error, then Error_CurrentError is set to that string
-->
<?LassoScript

	Define_Tag: '_SMTP_SendCommand', -Priority='Replace',
			-required='connection', -type='net',
			-optional='text',
			-optional='timeout',
			-optional='goodanswer',
			-optional='log';

		Local: '_timeout' = (integer: (local: 'timeout'));
		if: #_Timeout < 1 || #_Timeout > 120;
			local: '_timeout' = 15;
		/if;

		if: (local_defined: 'text') && (#text->size > 0);

			Local: 'myResult' = #connection->(wait: #_timeout, net_waitwrite);
			if: #myResult->(isa: 'integer') && (#myResult == net_waittimeout);
				Local: 'Answer'=(pair: '-8001'='SMTP server time out waiting for write');
				if: (local: 'log'); log_detail: 'ERROR ' + #answer; /if;
				return: #Answer;		
			/If;
	
			// Convert text to iso-8859-1 character set
			local:'bytes' = bytes;
			#bytes->(importAs: #text, 'iso8859-1');
			
			local: 'myResult' = #connection->(write: #bytes);
			if: (local: 'log'); log_detail: 'WRITE ' + #myResult + ' ' + #text; /if;
			
		/if;
		
		Local: 'myResult' = #connection->(wait: #_timeout, net_waitread);
		if: #myResult->(isa: 'integer') && (#myResult == net_waittimeout);
			Local: 'Answer'=(pair: '-8001'='SMTP server time out waiting for read');
			if: (local: 'log'); log_detail: 'ERROR ' + #answer; /if;
			return: #Answer;		
		/If;

		local: 'Answer' = #connection->(read: 32768);
		if: (local: 'log'); log_detail: 'READ ' + #answer; /if;
				
		If: Error_Code == 0;
			If: #Answer != '';
				Local: 'AnswerList'=(String_FindRegexp: #Answer, -Find='[0-9]+');
				error_seterrormessage: #Answer;
				If: #AnswerList->(get: 1) != (local: 'GoodAnswer');
					error_seterrorcode: (#AnswerList->get:1);
				Else;
					error_seterrorcode: 0;
				/If;
			Else;
				error_seterrormessage: 'SMTP server sent empty response';
				error_seterrrocode: -8001;
			/If;
		/If;
		Return: (pair: Error_Code=#Answer);
	/Define_Tag;
?>

<!----------------------------------------------------------------------->
<!-- _SMTP_SendOneMessage:
		Opens a SMTP dialog with a mail server and sends a single email message.
			Usually called by an external routine which is looping thru a list of messages
			to send.  It is the responsibility of the caller to retry on failure.

	Problems:
		Does not validate input, such as things like "no subject line", "no To: value", etc.


	Inputs:
		-Email.host			ip address or name of SMTP server
		-Email.from			email address of sender
		-Email.to			email address of recipient
		-Email.subject		(optional) subject line in outgoing email
		-Email.cc			(optional) email address of carbon-copy recipient
		-Email.bcc			(optional) email address of blind-carbon-copy recipient
		-Email.body		
		-Email.attachments	encoded text of message to be sent, including any attachments

	Outputs:
		Returns last message from mail server indicating success or failure

	Side-Effects:
		Sends an email out via the specified SMTP server
		Sets Error_CurrentError to last SMTP server message indicating failure
-->
[Define_Tag: '_SMTP_SendOneMessage', -Optional='Log', -Priority='Replace']

	[if: (local: 'log')][log_detail: 'SENDING ' + params][/if]

	[Protect]

		<!-- debug where crashed -->
		[local:'GotHere' = 0]

		<!-- init Answer to be NOT pair -->
		[local:'Answer' = '']

		<!-- identify Lasso crashes -->
		[handle: error_currenterror != error_noerror]
			[if: #Answer->type != 'pair']
				[if: local_defined:'message_chunk']
					[Local: 'Answer'=(pair: '-8001' = ('Unknown Bad Error ('+#GotHere+') : '+ (Error_CurrentError: -ErrorCode) + ' = ' + (Error_CurrentError: -ErrorMessage) + ' : Message_Chunk = ('+ (encode_stricturl: #message_chunk) +')' ) )]
					[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
					[return: #Answer]		
				[else]
					[Local: 'Answer'=(pair: '-8001' = ('Unknown Bad Error ('+#GotHere+') : '+ (Error_CurrentError: -ErrorCode) + ' = ' + (Error_CurrentError: -ErrorMessage) ) )]
					[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
					[return: #Answer]		
				[/if]
			[/if]
		[/handle]

		<!-- TCPIP connection timeout -->
		[Local: 'Timeout'=(integer: (_SMTP_GetPref: 'TCPTimeout'))]

		<!-- check for valid range -->
		[if: #Timeout < 1 || #Timeout > 120]
			[#Timeout = (integer: 15)]
		[/if]

		<!-- debug where crashed -->
		[local:'GotHere' = 1]

		<!-- get host and AUTH info, if given -->
		<!-- should be in format "username||password||smtp.host.domain" -->
		[local:'emailhost' = (local: 'Email.host')]
		[local:'authuser' = string]
		[local:'authpass' = string]
		[local:'auth' = false]
		[if: #emailhost->contains:'||']
			[local:'autharray' = #emailhost->split:'||']
			[if: #autharray->size == 3]
				[#authuser = encode_base64: (#autharray->get:1)]
				[#authpass = encode_base64: (#autharray->get:2)]	
				[#emailhost = #autharray->get:3]
				[#auth = true]
			[/if]
		[/if]

		<!-- Open a connection to SMTP server -->
		[Local: 'myConnection' = (Net)]
		[local: 'myResult' = #myConnection->(Connect: #emailhost, 25)]
		[if: (local: 'log')][log_detail: 'Connect ' + #myResult][/if]
		[if: #myResult != Net_ConnectOK]
			[Local: 'Answer'=(pair: '-8001'='Could not establish connection with SMTP server')]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]		
		[/If]
		[local: 'myResult' = #myConnection->(Wait: #timeout, Net_WaitRead)]
		[if: (local: 'log')][log_detail: 'Handshake ' + #myResult][/if]
		[if: #myResult->(isa: 'integer') && (#myResult == net_waittimeout)]
			[Local: 'Answer'=(pair: '-8001'='SMTP server time out waiting for handshake')]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]		
		[/If]
		
		<!-- debug where crashed -->
		[local:'GotHere' = 2]

		<!-- wait for first response "220 mail.server.com" from mail server -->
		[If: (Error_CurrentError: -ErrorCode) == 0]
			[Local: 'Answer'=(_SMTP_SendCommand:
				-Text='',
				-Timeout=#Timeout,
				-Log=(local: 'log'),
				-Connection=#myConnection,
				-GoodAnswer='220')]
		[Else]
			<!-- TCP error -->
			[Local: 'Answer'=(pair: (Error_CurrentError: -ErrorCode)=(Error_CurrentError: -ErrorMessage))]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]		
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 3]

		<!-- send "HELO [192.168.1.70]\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'Answer'=(_SMTP_SendCommand:
				-Text='HELO [' + $__lassoservice_ip__ + ']\r\n',
				-Timeout=#Timeout,
				-Log=(local: 'log'),
				-Connection=#myConnection,
				-GoodAnswer='250')]
		[else]
			[#Answer = (pair: (#answer->first) = 'Bad response from SMTP server on inital contact: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 4]

		<!-- send "RSET\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'Answer'=(_SMTP_SendCommand:
				-Text='RSET\r\n',
				-Timeout=#Timeout,
				-Log=(local: 'log'),
				-Connection=#myConnection,
				-GoodAnswer='250')]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to HELO: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- check if need to AUTH -->
		[if: #auth]
			<!-- send "AUTH LOGIN\r\n" -->
			[If: #Answer->first == 0]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text='AUTH LOGIN\r\n',
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='334')]
			[else]
				[#Answer = (pair: (#Answer->first)='Bad response to RSET: '+ (#Answer->second) )]
				[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
				[return: #Answer]
			[/If]

			<!-- send "username\r\n" -->
			[If: #Answer->first == 0]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text=#authuser + '\r\n',
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='334')]
			[else]
				[#Answer = (pair: (#Answer->first)='Bad response to AUTH: '+ (#Answer->second) )]
				[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
				[return: #Answer]
			[/If]

			<!-- send "password\r\n" -->
			[If: #Answer->first == 0]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text=#authpass + '\r\n',
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='235')]
			[else]
				[#Answer = (pair: (#Answer->first)='Bad response to your AUTH username: '+ (#Answer->second) )]
				[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
				[return: #Answer]
			[/If]

		[/if]

		<!-- debug where crashed -->
		[local:'GotHere' = 5]

		<!-- send "MAIL FROM:<GrantHulbert@blueworld.com>\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'FromList'=(_SMTP_FindEmails: (local: 'Email.from'))]
			[Loop: #FromList->size]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text=('MAIL FROM:<' + (#FromList->(get: Loop_Count)) + '>\r\n'),
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='250')]
			[/Loop]
		[else]
			[if: #auth]
				[#Answer = (pair: (#Answer->first)='Bad response to AUTH password: '+ (#Answer->second) )]
			[else]
				[#Answer = (pair: (#Answer->first)='Bad response to RSET: '+ (#Answer->second) )]
			[/if]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 6]

		<!-- Handle To:<blah> section: send "RCPT TO:<granthulbert@blueworld.com>\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'RecipientList'=(_SMTP_FindEmails: (local: 'Email.to'))]
			[Loop: #RecipientList->size]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text=('RCPT TO:<' + (#RecipientList->(get: Loop_Count)) + '>\r\n'),
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='250')]
			[/Loop]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to FROM: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 7]

		<!-- Handle CC:<blah> section: send "RCPT TO:<granthulbert@blueworld.com>\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'RecipientList'=(_SMTP_FindEmails: (local: 'Email.cc'))]
			[Loop: #RecipientList->size]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text=('RCPT TO:<' + (#RecipientList->(get: Loop_Count)) + '>\r\n'),
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='250')]
			[/Loop]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to TO: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 8]

		<!-- Handle BCC:<blah> section: send "RCPT TO:<granthulbert@blueworld.com>\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'RecipientList'=(_SMTP_FindEmails: (local: 'Email.bcc'))]
			[Loop: #RecipientList->size]
				[Local: 'Answer'=(_SMTP_SendCommand:
					-Text=('RCPT TO:<' + (#RecipientList->(get: Loop_Count)) + '>\r\n'),
					-Timeout=#Timeout,
					-Log=(local: 'log'),
					-Connection=#myConnection,
					-GoodAnswer='250')]
			[/Loop]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to CC: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 9]

		<!-- send "DATA\r\n" -->
		[If: #Answer->first == 0]
			[Local: 'Answer'=(_SMTP_SendCommand:
				-Text='DATA\r\n',
				-Timeout=#Timeout,
				-Log=(local: 'log'),
				-Connection=#myConnection,
				-GoodAnswer='354')]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to BCC: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 10]

		<!-- send message -->
		[If: #Answer->first == 0]
		
			[local:'_message'= (string: (local: 'Email.attachments'))->(split:'\r\n')]
			[iterate: #_message, local:'message_chunk']
				[local:'bytes' = (bytes)]
				[#bytes->(importAs: (#message_chunk + '\r\n'), 'iso8859-1')]
				[Local: 'myResult'=#myConnection->(write: #bytes)]
				[if: (local: 'log')][log_detail: 'DATA ' + #myResult + ' ' + #message_chunk][/if]
			[If: (Error_CurrentError: -ErrorCode) != 0]
				[loop_abort]
			[/If]
			[/iterate]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to DATA: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]
		
		<!-- debug where crashed -->
		[local:'GotHere' = 11]

		<!-- send "\r\n.\r\n" -->
		[If: (Error_CurrentError: -ErrorCode) == 0]
			[Local: 'Answer'=(_SMTP_SendCommand: -Text='\r\n.\r\n',
				-Timeout=#Timeout,
				-Log=(local: 'log'),
				-Connection=#myConnection,
				-GoodAnswer='250')]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to Sending DATA: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 12]

		<!-- send "QUIT\r\n" -->
		[If: (Error_CurrentError: -ErrorCode) == 0]
			[Local: 'Answer'=(_SMTP_SendCommand:
				-Text='QUIT\r\n',
				-Timeout=#Timeout,
				-Log=(local: 'log'),
				-Connection=#myConnection,
				-GoodAnswer='221')]
		[else]
			[#Answer = (pair: (#Answer->first)='Bad response to end of DATA: '+ (#Answer->second) )]
			[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
			[return: #Answer]
		[/If]

		<!-- debug where crashed -->
		[local:'GotHere' = 13]

		<!-- Close connection to SMTP server -->
		[handle] <!-- MAKE SURE IT ACTUALLY CLOSES NO MATTER WHAT!!! - KRJ -->
		[if: (local_defined: 'myconnection') && #myconnection->(isa: 'net')]
			[protect][#myconnection->close][/protect]
		[/if]

		[/handle]
		<!-- debug where crashed -->
		[local:'GotHere' = 14]

		<!-- Give the caller a nice string representing final disposition of SMTP dialog -->
			[if: (local: 'log')][log_detail: 'SUCCESS ' + #answer][/if]
		[return: #Answer]
	[/Protect]

	[if: (local_defined: 'myconnection') && #myconnection->(isa: 'net')]
		[protect][#myconnection->close][/protect]
	[/if]

	[if: (local: 'log')][log_detail: 'ERROR (' + (local: 'gothere') + ') ' + #answer][/if]
	[Return: #Answer]
[/Define_Tag]

<!--------------------------------------------------------------------->
<!-- _SMTP_SendOutgoingMail:
		Loops thru outgoing emails database and attempts to send all appropriate records

	Problems:
		retry seconds timeout is not read from prefs database, or implemented
		using made-up error of -8001
		not deleting attachments per prefs

	Inputs:
		None

	Outputs:
		None

	Side-Effects:
		Retrieves records from outgoing emails database, sends emails, modifies
		outgoing email records to indicate status of each sent/failed email
-->
[Define_Tag: '_SMTP_SendOutgoingMail', -Priority='Replace', -required='queue_number', -privileged]

	[Log_Warning: 'Email: Checking Email Queue #'+#queue_number]

	[Inline:	-Database='site',
			-Maxrecords='all',
			-SQL = 
				"SELECT * from _OutgoingEmail " +
				"WHERE " +
					"retries > 0 and " +
					"date_time_sent = '1000-01-01 00:00:00' and " +
					"(left(attachments,3) = 'To:' or " +
					"left(attachments,5) = 'Date:') and " +
					"ExtraMIME = " + #queue_number + " " +
				"LIMIT 100"
			]
	[if: found_count > 0]

		<!-- sleep each time through queue -->
		[sleep: 5000]

		[Records]
			[Protect]
				[Local: 'Answer'=(_SMTP_SendOneMessage:
					-Email.host=(Field: 'SMTPServer'),
					-Email.from=(Field: 'fromEmail'),
					-Email.to=(Field: 'toEmail'),
					-Email.subject=(Field: 'subject'),
					-Email.cc=(Field: 'cc'),
					-Email.bcc=(Field: 'bcc'),
					-Email.body=(Field: 'body'),
					-Email.attachments=(Field: 'attachments')
				)]
			[/Protect]

			[if: #Answer->type != 'pair']
				[#Answer = (pair: '-1'= 'Malformed Answer: ' + #Answer)]
			[/if]

			[Local: 'body'=(Field: 'body')]

			<!-- if successfully sent, deal with this record as the prefs suggest -->
			[If: (integer: #Answer->first) == 0]
				[log_detail: 'Email: Sent message to '+(field:'toEmail')+' from '+(field:'fromEmail')+' subject is '+(field:'subject')]
				
				<!-- if successfully sent, then modify the date_time_sent field and SMTPErr field -->
				[If: (Integer: (_SMTP_GetPref: 'RetainSuccessful')) == 1]
					<!-- if prefs say to blank out body, then do so -->
					[If: (Integer: (_SMTP_GetPref: 'RetainBody')) == 0]
						[Local: 'body'='']
					[/If]
					[Inline:  -Database='site', -Table='_OutgoingEmail', -Update,
						-KeyField='id',
						-KeyValue=(Field: 'id'),
						'date_time_sent'=(Date_Format: (Date_GetCurrentDate), -DateFormat='%Y-%m-%d %H:%M:%S'),
						'body'=#body,
						'SMTPErr'='Sent: ' + #Answer->second
					]
					[/Inline]
				[Else]
					<!-- if pref says to delete successful emails, then just delete it instead -->

					[Inline: -Database='site', -Delete, -Table='_OutgoingEmail',
						-KeyField='id', -KeyValue=(Field: 'id')
					]
					[/Inline]
				[/If]
			[Else]
				<!-- check to see if retry count has dropped to 0, in which case we delete as appropriate -->
				[If: (Integer: (Field: 'Retries')) == 1]

					<!-- If the retry count drops to 0, and pref says to delete failed emails, then delete it -->
					[If: (Integer: (_SMTP_GetPref: 'RetainError')) == 0]
						[Inline: -Database='site', -Delete, -Table='_OutgoingEmail',
							-KeyField='id', -KeyValue=(Field: 'id')
						]
						[/Inline]
					[Else]
						<!-- if not deleting failed emails, then insert an error message and stop retrying -->
						[Inline:  -Database='site', -Table='_OutgoingEmail', -Update,
							-KeyField='id',
							-KeyValue=(Field: 'id'),
							'Retries'='0',
							'SMTPErr'='Error Too many retries: ' + #Answer->first + ' = ' + #Answer->second
						]
						[/Inline]
					[/If]
				[Else]
					<!-- if not sent (and still retrying), then check to see if it's a recoverable or not -->
					[If: (string:#Answer->first)->beginswith:'5']
						<!-- fatal error, no more retries -->
						[Inline:  -Database='site', -Table='_OutgoingEmail', -Update,
							-KeyField='id',
							-KeyValue=(Field: 'id'),
							'Retries'='0',
							'SMTPErr'='Error: ' + #Answer->first + ' = ' + #Answer->second
						]
						[/Inline]					
					[Else]
						<!-- retry -->
						[Inline:  -Database='site', -Table='_OutgoingEmail', -Update,
							-KeyField='id',
							-KeyValue=(Field: 'id'),
							'Retries'=(Integer: (Field: 'Retries')) - 1,
							'SMTPErr'='Retry: ' + #Answer->first + ' = ' + #Answer->second
						]
						[/Inline]
					[/If]
				[/If]	<!-- retries 1 -->
			[/If]	<!-- failed email -->
		[/Records]
				
	[else]
		<!-- sleep for user-defined time -->
		[Sleep: (Integer: (_SMTP_GetPref: 'SleepSeconds')) * 1000]	
	[/if]

	[/Inline]
[/Define_Tag]


<!--------------------------------------------------------------------->
<!-- _SMTP_EmailWatcher:
		Loops forever in a separate thread looking for emails to send

	Problems:

	Inputs:
		None

	Outputs:
		None

	Side-Effects:
		calls _SMTP_SendOutgoingMail, which has side-effects
-->
[Define_Tag: '_SMTP_EmailWatcher', -async, -Priority='Replace']

	<!-- prevent average user from adding additional queues -->
	[auth_admin]

	<!-- get queue number -->
	[Local: 'queue_number' = (integer: $_SMTP_queue->find:'size') + 1]
	[$_SMTP_queue->(insert: 'size' = #queue_number)]

	[Log_Warning: 'Email: Starting Queue #'+#queue_number]
	[Sleep: 60000]
	[While: true]
		[Protect]
			[If: (Integer: (_SMTP_GetPref: 'Pause')) != 1]
				[_SMTP_SendOutgoingMail: #queue_number]
			[Else]
				[Log_Warning: 'Email: Queue #'+#queue_number+' Paused']
				[While: (Integer: (_SMTP_GetPref: 'Pause')) ]
					<!-- If email queue is paused, check for resume every 60 seconds -->
					[Sleep: 60000]	
				[/While]
				[Log_Warning: 'Email: Queue #'+#queue_number+' Unpaused']
			[/If]
		[/Protect]
	[/While]
	[Log_Warning: 'Email: Stopping Queue #'+#queue_number]
[/Define_Tag]
